from .frame_gen import ChipFrameGen, OfflineFrameGen, OnlineFrameGen
from .frame_gen_v2 import FrameGenV2, OfflineFrameGenV2
from .frames import (
    OfflineConfigFrame1,
    OfflineConfigFrame2,
    OfflineConfigFrame3,
    OfflineConfigFrame4,
    OfflineTestInFrame1,
    OfflineTestInFrame2,
    OfflineTestInFrame3,
    OfflineTestInFrame4,
    OfflineTestOutFrame1,
    OfflineTestOutFrame2,
    OfflineTestOutFrame3,
    OfflineTestOutFrame4,
    OfflineWorkFrame1,
    OfflineWorkFrame2,
    OfflineWorkFrame3,
    OfflineWorkFrame4,
    OnlineConfigFrame1,
    OnlineConfigFrame2,
    OnlineConfigFrame3,
    OnlineConfigFrame4,
    OnlineTestInFrame1,
    OnlineTestInFrame2,
    OnlineTestInFrame3,
    OnlineTestInFrame4,
    OnlineTestOutFrame1,
    OnlineTestOutFrame2,
    OnlineTestOutFrame3,
    OnlineTestOutFrame4,
    OnlineWorkFrame1_1,
    OnlineWorkFrame1_2,
    OnlineWorkFrame1_3,
    OnlineWorkFrame1_4,
    OnlineWorkFrame2,
    OnlineWorkFrame3,
    OnlineWorkFrame4,
)
from .types import (
    FRAME_DTYPE,
    LUT_DTYPE,
    PAYLOAD_DATA_DTYPE,
    FrameArrayType,
    LUTDataType,
    PayloadDataType,
    LUTPotentialType,
    LUT_POTENTIAL_DTYPE,
    LUTActivationType,
    LUT_ACTIVATION_DTYPE,
)

__all__ = [
    # Frame generators
    "OfflineFrameGen",
    "OnlineFrameGen",
    "ChipFrameGen",
    # Frame generators for v2.5
    "FrameGenV2",
    "OfflineFrameGenV2",
    # Classes of frames
    # Offline frames
    "OfflineConfigFrame1",
    "OfflineConfigFrame2",
    "OfflineConfigFrame3",
    "OfflineConfigFrame4",
    "OfflineWorkFrame1",
    "OfflineWorkFrame2",
    "OfflineWorkFrame3",
    "OfflineWorkFrame4",
    "OfflineTestInFrame1",
    "OfflineTestInFrame2",
    "OfflineTestInFrame3",
    "OfflineTestInFrame4",
    "OfflineTestOutFrame1",
    "OfflineTestOutFrame2",
    "OfflineTestOutFrame3",
    "OfflineTestOutFrame4",
    # Online frames
    "OnlineConfigFrame1",
    "OnlineConfigFrame2",
    "OnlineConfigFrame3",
    "OnlineConfigFrame4",
    "OnlineWorkFrame1_1",
    "OnlineWorkFrame1_2",
    "OnlineWorkFrame1_3",
    "OnlineWorkFrame1_4",
    "OnlineWorkFrame2",
    "OnlineWorkFrame3",
    "OnlineWorkFrame4",
    "OnlineTestInFrame1",
    "OnlineTestInFrame2",
    "OnlineTestInFrame3",
    "OnlineTestInFrame4",
    "OnlineTestOutFrame1",
    "OnlineTestOutFrame2",
    "OnlineTestOutFrame3",
    "OnlineTestOutFrame4",
    # Common Types
    "FRAME_DTYPE",
    "LUT_DTYPE",
    "PAYLOAD_DATA_DTYPE",
    "FrameArrayType",
    "LUTDataType",
    "PayloadDataType",
    "LUTPotentialType",
    "LUT_POTENTIAL_DTYPE",
    "LUTActivationType",
    "LUT_ACTIVATION_DTYPE",
]



from enum import IntEnum, unique

from ..utils import _mask


@unique
class FrameType(IntEnum):
    """Basic types of Frames"""

    CONFIG = 0
    TEST = 0x1
    WORK = 0x2
    CTRL = 0x3


FT = FrameType


@unique
class FrameHeader(IntEnum):
    """Frame headers"""

    CONFIG_TYPE1 = (FT.CONFIG << 2) | 0b00
    CONFIG_TYPE2 = (FT.CONFIG << 2) | 0b01
    CONFIG_TYPE3 = (FT.CONFIG << 2) | 0b10
    CONFIG_TYPE4 = (FT.CONFIG << 2) | 0b11

    TEST_TYPE1 = (FT.TEST << 2) | 0b00
    TEST_TYPE2 = (FT.TEST << 2) | 0b01
    TEST_TYPE3 = (FT.TEST << 2) | 0b10
    TEST_TYPE4 = (FT.TEST << 2) | 0b11

    WORK_TYPE1 = (FT.WORK << 2) | 0b00
    WORK_TYPE2 = (FT.WORK << 2) | 0b01
    WORK_TYPE3 = (FT.WORK << 2) | 0b10
    WORK_TYPE4 = (FT.WORK << 2) | 0b11

    CTRL_TYPE1 = (FT.CTRL << 2) | 0b00
    CTRL_TYPE2 = (FT.CTRL << 2) | 0b01
    CTRL_TYPE3 = (FT.CTRL << 2) | 0b10
    CTRL_TYPE4 = (FT.CTRL << 2) | 0b11


@unique
class FramePackageType(IntEnum):
    """Frame package type."""

    CONF_TESTOUT = 0  # For conf & test-out
    TESTIN = 1  # For test-in


class FrameFormat:
    """General frame mask & offset."""

    FRAME_LENGTH = 64
    GENERAL_MASK = _mask(FRAME_LENGTH)

    # Header
    GENERAL_HEADER_OFFSET = 60
    GENERAL_HEADER_MASK = _mask(4)

    GENERAL_FRAME_TYPE_OFFSET = GENERAL_HEADER_OFFSET
    GENERAL_FRAME_TYPE_MASK = GENERAL_HEADER_MASK

    # Chip address
    GENERAL_CHIP_ADDR_OFFSET = 50
    GENERAL_CHIP_ADDR_MASK = _mask(10)
    # Chip X/Y address
    GENERAL_CHIP_X_ADDR_OFFSET = 55
    GENERAL_CHIP_X_ADDR_MASK = _mask(5)
    GENERAL_CHIP_Y_ADDR_OFFSET = GENERAL_CHIP_ADDR_OFFSET
    GENERAL_CHIP_Y_ADDR_MASK = _mask(5)

    # Core address
    GENERAL_CORE_ADDR_OFFSET = 40
    GENERAL_CORE_ADDR_MASK = _mask(10)
    # Core X/Y address
    GENERAL_CORE_X_ADDR_OFFSET = 45
    GENERAL_CORE_X_ADDR_MASK = _mask(5)
    GENERAL_CORE_Y_ADDR_OFFSET = GENERAL_CORE_ADDR_OFFSET
    GENERAL_CORE_Y_ADDR_MASK = _mask(5)

    # Core* address
    GENERAL_CORE_EX_ADDR_OFFSET = 30
    GENERAL_CORE_EX_ADDR_MASK = _mask(10)
    # Core* X/Y address
    GENERAL_CORE_X_EX_ADDR_OFFSET = 35
    GENERAL_CORE_X_EX_ADDR_MASK = _mask(5)
    GENERAL_CORE_Y_EX_ADDR_OFFSET = GENERAL_CORE_EX_ADDR_OFFSET
    GENERAL_CORE_Y_EX_ADDR_MASK = _mask(5)

    # Global core = Chip address + core address
    GENERAL_CORE_GLOBAL_ADDR_OFFSET = GENERAL_CORE_ADDR_OFFSET
    GENERAL_CORE_GLOBAL_ADDR_MASK = _mask(20)

    # Payload
    GENERAL_PAYLOAD_OFFSET = 0
    GENERAL_PAYLOAD_LENGTH = 30
    GENERAL_PAYLOAD_MASK = _mask(GENERAL_PAYLOAD_LENGTH)

    # Package neuron start address
    # NOTE: The argument `sram_base_addr` of config/test frame 3 & 4 is incorrectly
    # named. In fact, it is the neuron start address.
    GENERAL_PACKAGE_NEU_START_ADDR_OFFSET = 20
    GENERAL_PACKAGE_NEU_START_ADDR_MASK = _mask(10)

    # Package type
    GENERAL_PACKAGE_TYPE_OFFSET = 19
    GENERAL_PACKAGE_TYPE_MASK = _mask(1)

    # #N of packages
    GENERAL_PACKAGE_NUM_OFFSET = 0
    GENERAL_PACKAGE_NUM_MASK = _mask(19)

    # Package attributes
    GENERAL_PACKAGE_LEN = FRAME_LENGTH


FF = FrameFormat


class _TestFrameFormat_In(FF):
    """General test input frame format."""

    pass


class _TestFrameFormat_Out(FF):
    """General test output frame format."""

    TEST_CHIP_ADDR_OFFSET = FF.GENERAL_CHIP_ADDR_OFFSET
    TEST_CHIP_ADDR_MASK = FF.GENERAL_CHIP_ADDR_OFFSET


"""Frame format for offline cores"""


class OfflineConfigFrame1Format(FF):
    """Offline config frame type I. Random seed register."""

    pass


class OfflineConfigFrame2Format(FF):
    """Offline config frame type II. Parameter registers."""

    # Frame #1
    WEIGHT_WIDTH_OFFSET = 28
    WEIGHT_WIDTH_MASK = _mask(2)

    LCN_OFFSET = 24
    LCN_MASK = _mask(4)

    INPUT_WIDTH_OFFSET = 23
    INPUT_WIDTH_MASK = 1

    SPIKE_WIDTH_OFFSET = 22
    SPIKE_WIDTH_MASK = 1

    # NOTE: This parameter actually means the effective number of dendrites,
    # which can be found in the manual for details.
    NUM_DENDRITE_OFFSET = 9
    NUM_DENDRITE_MASK = _mask(13)

    MAX_POOLING_EN_OFFSET = 8
    MAX_POOLING_EN_MASK = _mask(1)

    TICK_WAIT_START_HIGH8_OFFSET = 0
    TICK_WAIT_START_HIGH8_MASK = _mask(8)

    # Frame #2
    TICK_WAIT_START_LOW7_OFFSET = 23
    TICK_WAIT_START_LOW7_MASK = _mask(7)

    TICK_WAIT_END_OFFSET = 8
    TICK_WAIT_END_MASK = _mask(15)

    SNN_EN_OFFSET = 7
    SNN_EN_MASK = _mask(1)

    TARGET_LCN_OFFSET = 3
    TARGET_LCN_MASK = _mask(4)

    TEST_CHIP_ADDR_HIGH3_OFFSET = 0
    TEST_CHIP_ADDR_HIGH3_MASK = _mask(3)

    # Frame #3
    TEST_CHIP_ADDR_LOW7_OFFSET = 23
    TEST_CHIP_ADDR_LOW7_MASK = _mask(7)


class OfflineConfigFrame3Format(FF):
    """Offline config frame type III. Neuron RAM.

    Total payload: 256 bits, LSB.
    Frame #1: RAM[63:0]
    Frame #2: RAM[127:64]
    Frame #3: RAM[191:128]
    Frame #4: 42'd0, RAM[213:192]
    """

    # Frame #1
    VOLTAGE_OFFSET = 0
    VOLTAGE_MASK = _mask(30)

    BIT_TRUNC_OFFSET = 30
    BIT_TRUNC_MASK = _mask(5)

    SYN_INTEGRATION_MODE_OFFSET = 35
    SYN_INTEGRATION_MODE_MASK = _mask(1)

    LEAK_V_LOW28_OFFSET = 36
    LEAK_V_LOW28_MASK = _mask(28)

    # Frame #2
    LEAK_V_HIGH2_OFFSET = 0
    LEAK_V_HIGH2_MASK = _mask(2)

    LEAK_INTEGRATION_MODE_OFFSET = 2
    LEAK_INTEGRATION_MODE_MASK = _mask(1)

    LEAK_DIRECTION_OFFSET = 3
    LEAK_DIRECTION_MASK = _mask(1)

    POS_THRESHOLD_OFFSET = 4
    POS_THRESHOLD_MASK = _mask(29)

    NEG_THRES_OFFSET = 33
    NEG_THRES_MASK = _mask(29)

    NEG_THRES_MODE_OFFSET = 62
    NEG_THRES_MODE_MASK = _mask(1)

    THRE_MASK_BITS_LOW1_OFFSET = 63
    THRE_MASK_BITS_LOW1_MASK = _mask(1)

    # Frame #3
    THRES_MASK_BITS_HIGH4_OFFSET = 0
    THRES_MASK_BITS_HIGH4_MASK = _mask(4)

    LEAK_COMPARISON_OFFSET = 4
    LEAK_COMPARISON_MASK = _mask(1)

    RESET_V_OFFSET = 5
    RESET_V_MASK = _mask(30)

    RESET_MODE_OFFSET = 35
    RESET_MODE_MASK = _mask(2)

    ADDR_CHIP_Y_OFFSET = 37
    ADDR_CHIP_Y_MASK = _mask(5)

    ADDR_CHIP_X_OFFSET = 42
    ADDR_CHIP_X_MASK = _mask(5)

    ADDR_CORE_Y_EX_OFFSET = 47
    ADDR_CORE_Y_EX_MASK = _mask(5)

    ADDR_CORE_X_EX_OFFSET = 52
    ADDR_CORE_X_EX_MASK = _mask(5)

    ADDR_CORE_Y_OFFSET = 57
    ADDR_CORE_Y_MASK = _mask(5)

    ADDR_CORE_X_LOW2_OFFSET = 62
    ADDR_CORE_X_LOW2_MASK = _mask(2)

    # Frame #4
    ADDR_CORE_X_HIGH3_OFFSET = 0
    ADDR_CORE_X_HIGH3_MASK = _mask(3)

    ADDR_AXON_OFFSET = 3
    ADDR_AXON_MASK = _mask(11)

    TICK_RELATIVE_OFFSET = 14
    TICK_RELATIVE_MASK = _mask(8)


class OfflineConfigFrame4Format(FF):
    """Offline config frame type IV. Weight RAM."""

    pass


OfflineRandomSeedFormat = OfflineConfigFrame1Format
OfflineCoreRegFormat = OfflineConfigFrame2Format
OfflineNeuRAMFormat = OfflineConfigFrame3Format
OfflineWeightRAMFormat = OfflineConfigFrame4Format


class OfflineWorkFrame1Format(FF):
    """Work frame type I. Spike."""

    RESERVED_OFFSET = 27
    RESERVED_MASK = _mask(3)

    AXON_OFFSET = 16
    AXON_MASK = _mask(11)

    TIMESLOT_OFFSET = 8
    TIMESLOT_MASK = _mask(8)

    DATA_OFFSET = 0
    DATA_MASK = _mask(8)


class OfflineWorkFrame2Format(FF):
    """Work frame type II. Sync."""

    RESERVED_OFFSET = 30
    RESERVED_MASK = _mask(20)

    # #N of sync signals
    N_SYNC_OFFSET = 0
    N_SYNC_MASK = _mask(30)


class OfflineWorkFrame3Format(FF):
    """Work frame type III. Clear."""

    RESERVED_OFFSET = 0
    RESERVED_MASK = _mask(50)


class OfflineWorkFrame4Format(FF):
    """Work frame type IV. Init."""

    RESERVED_OFFSET = 0
    RESERVED_MASK = _mask(50)


SpikeFrameFormat = OfflineWorkFrame1Format
SyncFrameFormat = OfflineWorkFrame2Format
ClearFrameFormat = OfflineWorkFrame3Format
InitFrameFormat = OfflineWorkFrame4Format


class _OfflineTestFrameFormat_In(_TestFrameFormat_In):
    """General offline test input frame format."""

    pass


class _OfflineTestFrameFormat_Out(_TestFrameFormat_Out):
    """General offline test output frame format."""

    pass


class OfflineTestFrame1Format_In(_OfflineTestFrameFormat_In):
    """Test frame type I. Random seed register, input."""

    pass


class OfflineTestFrame2Format_In(_OfflineTestFrameFormat_In):
    """Test frame type II. Parameter registers, input."""

    pass


class OfflineTestFrame3Format_In(_OfflineTestFrameFormat_In):
    """Test frame type III. Neuron RAM, input."""

    pass


class OfflineTestFrame4Format_In(_OfflineTestFrameFormat_In):
    """Test frame type IV. Weight RAM, input."""

    pass


class OfflineTestFrame1Format_Out(
    _OfflineTestFrameFormat_Out, OfflineConfigFrame1Format
):
    """Test frame type I. Random seed register, output."""

    pass


class OfflineTestFrame2Format_Out(
    _OfflineTestFrameFormat_Out, OfflineConfigFrame2Format
):
    """Test frame type II. Parameter registers, output."""

    pass


class OfflineTestFrame3Format_Out(
    _OfflineTestFrameFormat_Out, OfflineConfigFrame3Format
):
    """Test frame type III. Neuron RAM, output."""

    pass


class OfflineTestFrame4Format_Out(
    _OfflineTestFrameFormat_Out, OfflineConfigFrame4Format
):
    """Test frame type IV. Weight RAM, output."""

    pass


"""Frame format for online cores"""


class OnlineConfigFrame1Format(FF):
    """Online config frame type I. LUT."""

    LUT_OFFSET = 0
    LUT_MASK = _mask(30)


class OnlineConfigFrame2Format(FF):
    """Online config frame type II. Core."""

    # Frame #1
    WEIGHT_WIDTH_OFFSET = 28
    WEIGHT_WIDTH_MASK = _mask(2)

    LCN_OFFSET = 26
    LCN_MASK = _mask(2)

    LATERAL_INHI_VALUE_HIGH26_OFFSET = 0
    LATERAL_INHI_VALUE_HIGH26_MASK = _mask(26)

    # Frame #2
    LATERAL_INHI_VALUE_LOW6_OFFSET = 24
    LATERAL_INHI_VALUE_LOW6_MASK = _mask(6)

    WEIGHT_DECAY_OFFSET = 16
    WEIGHT_DECAY_MASK = _mask(8)

    UPPER_WEIGHT_OFFSET = 8
    UPPER_WEIGHT_MASK = _mask(8)

    LOWER_WEIGHT_OFFSET = 0
    LOWER_WEIGHT_MASK = _mask(8)

    # Frame #3
    NEURON_START_OFFSET = 20
    NEURON_START_MASK = _mask(10)

    NEURON_END_OFFSET = 10
    NEURON_END_MASK = _mask(10)

    INHI_CORE_X_EX_OFFSET = 5
    INHI_CORE_X_EX_MASK = _mask(5)

    INHI_CORE_Y_EX_OFFSET = 0
    INHI_CORE_Y_EX_MASK = _mask(5)

    # Frame #4
    TICK_WAIT_START_OFFSET = 15
    TICK_WAIT_START_MASK = _mask(15)

    TICK_WAIT_END_OFFSET = 0
    TICK_WAIT_END_MASK = _mask(15)

    # Frame #5
    LUT_RANDOM_EN_HIGH30_OFFSET = 0
    LUT_RANDOM_EN_HIGH30_MASK = _mask(30)

    # Frame #6
    LUT_RANDOM_EN_LOW30_OFFSET = 30
    LUT_RANDOM_EN_LOW30_MASK = _mask(30)

    # Frame #7
    DECAY_RANDOM_EN_OFFSET = 29
    DECAY_RANDOM_EN_MASK = _mask(1)

    LEAK_ORDER_OFFSET = 27
    LEAK_ORDER_MASK = _mask(1)

    ONLINE_MODE_EN_OFFSET = 26
    ONLINE_MODE_EN_MASK = _mask(1)

    TEST_CHIP_ADDR_OFFSET = 16
    TEST_CHIP_ADDR_MASK = _mask(10)

    RANDOM_SEED_OFFSET = 0
    RANDOM_SEED_MASK = _mask(16)

    # Frame #8
    ZERO_PADDING_OFFSET = 0
    ZERO_PADDING_MASK = _mask(30)


class OnlineConfigFrame3Format(FF):
    """Online config frame type III. Neuron RAM.

    Total payload: 256 bits, MSB.
    Frame #1: RAM[127:64]
    Frame #2: RAM[63:0]

    NOTE: For neuron in cores with 1-bit weight width: use 2 frame only.
          For neuron in cores with other weight width: use 4 frames.
    """

    # Frame #1, common for WW=1 & WW2/4/8
    PLASTICITY_END_OFFSET = 0
    PLASTICITY_END_MASK = _mask(10)

    PLASTICITY_START_OFFSET = 10
    PLASTICITY_START_MASK = _mask(10)

    ADDR_AXON_OFFSET = 20
    ADDR_AXON_MASK = _mask(11)

    ADDR_CORE_Y_EX_OFFSET = 31
    ADDR_CORE_Y_EX_MASK = _mask(5)

    ADDR_CORE_X_EX_OFFSET = 36
    ADDR_CORE_X_EX_MASK = _mask(5)

    ADDR_CORE_Y_OFFSET = 41
    ADDR_CORE_Y_MASK = _mask(5)

    ADDR_CORE_X_OFFSET = 46
    ADDR_CORE_X_MASK = _mask(5)

    ADDR_CHIP_Y_OFFSET = 51
    ADDR_CHIP_Y_MASK = _mask(5)

    ADDR_CHIP_X_OFFSET = 56
    ADDR_CHIP_X_MASK = _mask(5)

    TIME_RELATIVE_OFFSET = 61
    TIME_RELATIVE_MASK = _mask(3)


class OnlineConfigFrame3Format_WW1(OnlineConfigFrame3Format):
    """Online config frame type III with 1-bit weight width."""

    # Frame #1
    LEAK_V_OFFSET = 49
    LEAK_V_MASK = _mask(15)

    POS_THRES_OFFSET = 34
    POS_THRES_MASK = _mask(15)

    NEG_THRES_OFFSET = 27
    NEG_THRES_MASK = _mask(7)

    RESET_V_OFFSET = 21
    RESET_V_MASK = _mask(6)

    INIT_V_OFFSET = 15
    INIT_V_MASK = _mask(6)

    VOLTAGE_OFFSET = 0
    VOLTAGE_MASK = _mask(15)


class OnlineConfigFrame3Format_WWn(OnlineConfigFrame3Format):
    """Online config frame type III with 2/4/8-bit weight width."""

    # Frame #1
    LEAK_V_OFFSET = 32
    LEAK_V_MASK = _mask(32)

    POS_THRES_OFFSET = 0
    POS_THRES_MASK = _mask(32)

    # Frame #2
    NEG_THRES_OFFSET = 32
    NEG_THRES_MASK = _mask(32)

    RESET_V_OFFSET = 0
    RESET_V_MASK = _mask(32)

    # Frame #3
    INIT_V_OFFSET = 32
    INIT_V_MASK = _mask(32)

    VOLTAGE_OFFSET = 32
    VOLTAGE_MASK = _mask(0)


class OnlineConfigFrame4Format(FF):
    """Online config frame type IV. Weight RAM."""

    pass


OnlineLUTFormat = OnlineConfigFrame1Format
OnlineCoreRegFormat = OnlineConfigFrame2Format
OnlineNeuRAMFormat = OnlineConfigFrame3Format
OnlineNeuRAMFormat_WW1 = OnlineConfigFrame3Format_WW1
OnlineNeuRAMFormat_WWn = OnlineConfigFrame3Format_WWn
OnlineWeightRAMFormat = OnlineConfigFrame4Format


@unique
class Online_WF1F_SubType(IntEnum):
    TYPE_I = 0b000
    TYPE_II = 0b100
    TYPE_III = 0b010
    TYPE_IV = 0b001


class OnlineWorkFrame1Format(FF):
    """Work frame type I."""

    SUBTYPE_OFFSET = 27
    SUBTYPE_MASK = _mask(3)


class OnlineWorkFrame1Format_1(OnlineWorkFrame1Format):
    """Work frame type I, sub-type I. Spike."""

    AXON_OFFSET = 16
    AXON_MASK = _mask(11)

    RESERVED_1_OFFSET = 11
    RESERVED_1_MASK = _mask(5)

    TIMESLOT_OFFSET = 8
    TIMESLOT_MASK = _mask(3)

    RESERVED_2_OFFSET = 0
    RESERVED_2_MASK = _mask(8)


class OnlineWorkFrame1Format_2(OnlineWorkFrame1Format):
    """Work frame type I, sub-type II. Start."""

    RESERVED_OFFSET = 0
    RESERVED_MASK = _mask(27)


class OnlineWorkFrame1Format_3(OnlineWorkFrame1Format):
    """Work frame type I, sub-type III. End."""

    RESERVED_OFFSET = 0
    RESERVED_MASK = _mask(27)


class OnlineWorkFrame1Format_4(OnlineWorkFrame1Format):
    """Work frame type I, sub-type IV. Lateral inhibition."""

    RESERVED_OFFSET = 0
    RESERVED_MASK = _mask(27)


class OnlineWorkFrame2Format(OfflineWorkFrame2Format):
    """Work frame type II. Sync."""


class OnlineWorkFrame3Format(OfflineWorkFrame3Format):
    """Work frame type III. Clear."""


class OnlineWorkFrame4Format(OfflineWorkFrame4Format):
    """Work frame type IV. Init."""


class _OnlineTestFrameFormat_In(_TestFrameFormat_In):
    """General online test input frame format."""

    pass


class _OnlineTestFrameFormat_Out(_TestFrameFormat_Out):
    """General online test output frame format."""

    pass


class OnlineTestFrame1Format_In(_OnlineTestFrameFormat_In):
    """Test frame type I. LUT, input."""

    pass


class OnlineTestFrame2Format_In(_OnlineTestFrameFormat_In):
    """Test frame type II. Core, input."""

    pass


class OnlineTestFrame3Format_In(_OnlineTestFrameFormat_In):
    """Test frame type III. Neuron RAM, input."""

    pass


class OnlineTestFrame4Format_In(_OnlineTestFrameFormat_In):
    """Test frame type IV. Weight RAM, input."""

    pass


class OnlineTestFrame1Format_Out(_OnlineTestFrameFormat_Out, OnlineConfigFrame1Format):
    """Test frame type I. LUT, output."""

    pass


class OnlineTestFrame2Format_Out(_OnlineTestFrameFormat_Out, OnlineConfigFrame2Format):
    """Test frame type II. Core, output."""

    pass


class OnlineTestFrame3Format_Out(_OnlineTestFrameFormat_Out, OnlineConfigFrame3Format):
    """Test frame type III. Neuron RAM, output."""

    pass


class OnlineTestFrame4Format_Out(_OnlineTestFrameFormat_Out, OnlineConfigFrame4Format):
    """Test frame type IV. Weight RAM, output."""

    pass


class FrameFormatV2:
    """General frame mask & offset."""

    FRAME_LENGTH = 64
    GENERAL_MASK = _mask(FRAME_LENGTH)

    # Header
    GENERAL_HEADER_OFFSET = 60
    GENERAL_HEADER_MASK = _mask(4)

    GENERAL_FRAME_TYPE_OFFSET = GENERAL_HEADER_OFFSET
    GENERAL_FRAME_TYPE_MASK = GENERAL_HEADER_MASK

    # Core address
    GENERAL_CORE_ADDR_OFFSET = 42
    GENERAL_CORE_ADDR_MASK = _mask(18)
    # Core XY address
    GENERAL_CORE_XY_ADDR_OFFSET = 54
    GENERAL_CORE_XY_ADDR_MASK = _mask(6)
    # Core X address
    GENERAL_CORE_X_ADDR_OFFSET = 48
    GENERAL_CORE_X_ADDR_MASK = _mask(6)
    # Core Y address
    GENERAL_CORE_Y_ADDR_OFFSET = GENERAL_CORE_ADDR_OFFSET
    GENERAL_CORE_Y_ADDR_MASK = _mask(6)

    # Copy address
    GENERAL_COPY_ADDR_OFFSET = 24
    GENERAL_COPY_ADDR_MASK = _mask(18)
    # Copy XY address
    GENERAL_COPY_XY_ADDR_OFFSET = 36
    GENERAL_COPY_XY_ADDR_MASK = _mask(6)
    # Copy X address
    GENERAL_COPY_X_ADDR_OFFSET = 30
    GENERAL_COPY_X_ADDR_MASK = _mask(6)
    # Copy Y address
    GENERAL_COPY_Y_ADDR_OFFSET = GENERAL_COPY_ADDR_OFFSET
    GENERAL_COPY_Y_ADDR_MASK = _mask(6)

    # Payload
    GENERAL_PAYLOAD_OFFSET = 0
    GENERAL_PAYLOAD_LENGTH = 24
    GENERAL_PAYLOAD_MASK = _mask(GENERAL_PAYLOAD_LENGTH)

    # Package type
    GENERAL_PACKAGE_TYPE_OFFSET = 23
    GENERAL_PACKAGE_TYPE_MASK = _mask(1)

    # Package neuron start address
    GENERAL_PACKAGE_NEU_START_ADDR_OFFSET = 14
    GENERAL_PACKAGE_NEU_START_ADDR_MASK = _mask(9)

    # #N of packages
    GENERAL_PACKAGE_NUM_OFFSET = 0
    GENERAL_PACKAGE_NUM_MASK = _mask(14)

    # Package attributes
    GENERAL_PACKAGE_LEN = FRAME_LENGTH


FFV2 = FrameFormatV2


class _TestFrameFormat_InV2(FFV2):
    """General test input frame format."""

    pass


class _TestFrameFormat_OutV2(FFV2):
    """General test output frame format."""

    pass


class OfflineConfigFrame1FormatV2(FFV2):
    """Offline config frame type I. Core Parameter.

    Structure:
        Word0: Frame #1 (Bits 191-128)
        Word1: Frame #2 (Bits 127-64)
        Word2: Frame #3 (Bits 63-0)
    """

    class Word1:
        """Frame #1"""

        SNN_ANN_OFFSET = 63
        SNN_ANN_MASK = _mask(1)

        MAX_POOLING_OFFSET = 62
        MAX_POOLING_MASK = _mask(1)

        ADD_POTENTIAL_OFFSET = 61
        ADD_POTENTIAL_MASK = _mask(1)

        ZERO_OUTPUT_OFFSET = 60
        ZERO_OUTPUT_MASK = _mask(1)

        INPUT_SIGN_OFFSET = 59
        INPUT_SIGN_MASK = _mask(1)

        INPUT_WIDTH_OFFSET = 57
        INPUT_WIDTH_MASK = _mask(2)

        OUTPUT_SIGN_OFFSET = 56
        OUTPUT_SIGN_MASK = _mask(1)

        OUTPUT_WIDTH_OFFSET = 54
        OUTPUT_WIDTH_MASK = _mask(2)

        WEIGHT_SIGN_OFFSET = 53
        WEIGHT_SIGN_MASK = _mask(1)

        WEIGHT_WIDTH_OFFSET = 51
        WEIGHT_WIDTH_MASK = _mask(2)

        LCN_OFFSET = 47
        LCN_MASK = _mask(4)

        TARGET_LCN_OFFSET = 43
        TARGET_LCN_MASK = _mask(4)

        AXON_SKEW_OFFSET = 27
        AXON_SKEW_MASK = _mask(16)

        NEURON_NUMBER_OFFSET = 14
        NEURON_NUMBER_MASK = _mask(13)

        TEST_CORE_XY_OFFSET = 8
        TEST_CORE_XY_MASK = _mask(6)

        TEST_CORE_X_OFFSET = 2
        TEST_CORE_X_MASK = _mask(6)

        TEST_CORE_Y_HIGH2_OFFSET = 0
        TEST_CORE_Y_HIGH2_MASK = _mask(2)

    class Word2:
        """Frame #2"""

        TEST_CORE_Y_LOW4_OFFSET = 60
        TEST_CORE_Y_LOW4_MASK = _mask(4)

        GLOBAL_SEND_OFFSET = 53
        GLOBAL_SEND_MASK = _mask(7)

        CSC_ACCELERATE_OFFSET = 52
        CSC_ACCELERATE_MASK = _mask(1)

        GLOBAL_RECEIVE_OFFSET = 46
        GLOBAL_RECEIVE_MASK = _mask(6)

        THREAD_NUMBER_OFFSET = 36
        THREAD_NUMBER_MASK = _mask(10)

        BUSY_CYCLE_OFFSET = 24
        BUSY_CYCLE_MASK = _mask(12)

        DELAY_CYCLE_OFFSET = 8
        DELAY_CYCLE_MASK = _mask(16)

        WIDTH_CYCLE_OFFSET = 0
        WIDTH_CYCLE_MASK = _mask(8)

    class Word3:
        """Frame #3"""

        TICK_START_OFFSET = 48
        TICK_START_MASK = _mask(16)

        TICK_DURATION_OFFSET = 16
        TICK_DURATION_MASK = _mask(32)

        TICK_INITIAL_OFFSET = 0
        TICK_INITIAL_MASK = _mask(16)


class OfflineConfigFrame2FormatV2(FFV2):
    """Offline config frame type II. LUT SRAM.

    RAM_0 256x32 bit RMA_1 256x8 bit
    Frame #1: 24'd0 + RAM_0[0] + RAM_1[0]
    Frame #2: 24'd0 + RAM_0[1] + RAM_1[1]
    Frame #3: 24'd0 + RAM_0[2] + RAM_1[2]
    ...
    Frame #256: 24'd0 + RAM_0[255] + RAM_1[255]
    """

    POTENTIAL_OFFSET = 8
    POTENTIAL_MASK = _mask(32)

    ACTIVATION_OFFSET = 0
    ACTIVATION_MASK = _mask(8)


class OfflineConfigFrame3FormatV2(FFV2):
    """Offline config frame type III. Neuron SRAM.
    RAM 4096x128 bit
    Frame #1: RAM[0][63:0]
    Frame #2: RAM[0][127:64]
    Frame #3: RAM[1][63:0]
    ...
    Frame #8191: RAM[4095][63:0]
    Frame #8192: RAM[4095][127:64]
    """

    class Full:
        class Word1:
            WEIGHT_SKEW_LOW5_OFFSET = 59
            WEIGHT_SKEW_LOW5_MASK = _mask(5)

            WEIGHT_ADDRESS_START_OFFSET = 47
            WEIGHT_ADDRESS_START_MASK = _mask(12)

            WEIGHT_ADDRESS_END_OFFSET = 35
            WEIGHT_ADDRESS_END_MASK = _mask(12)

            OUTPUT_TYPE_OFFSET = 34
            OUTPUT_TYPE_MASK = _mask(1)

            FOLD_TYPE_OFFSET = 33
            FOLD_TYPE_MASK = _mask(1)

            NEURON_TYPE_OFFSET = 32
            NEURON_TYPE_MASK = _mask(1)

            VJT_OFFSET = 0
            VJT_MASK = _mask(32)

        class Word2:
            TICK_RELATIVE_OFFSET = 56
            TICK_RELATIVE_MASK = _mask(8)

            ADDR_AXON_OFFSET = 47
            ADDR_AXON_MASK = _mask(9)

            ADDR_CORE_XY_OFFSET = 41
            ADDR_CORE_XY_MASK = _mask(6)

            ADDR_CORE_X_OFFSET = 35
            ADDR_CORE_X_MASK = _mask(6)

            ADDR_CORE_Y_OFFSET = 29
            ADDR_CORE_Y_MASK = _mask(6)

            ADDR_COPY_XY_OFFSET = 23
            ADDR_COPY_XY_MASK = _mask(6)

            ADDR_COPY_X_OFFSET = 17
            ADDR_COPY_X_MASK = _mask(6)

            ADDR_COPY_Y_OFFSET = 11
            ADDR_COPY_Y_MASK = _mask(6)

            WEIGHT_SKEW_HIGH11_OFFSET = 0
            WEIGHT_SKEW_HIGH11_MASK = _mask(11)

        class Word3:
            THRESHOLD_POS_LOW20_OFFSET = 44
            THRESHOLD_POS_LOW20_MASK = _mask(20)

            LATERAL_INHIBITION_OFFSET = 43
            LATERAL_INHIBITION_MASK = _mask(1)

            LEAK_MULTI_SEQUENCE_OFFSET = 42
            LEAK_MULTI_SEQUENCE_MASK = _mask(1)

            LEAK_MULTI_INPUT_OFFSET = 41
            LEAK_MULTI_INPUT_MASK = _mask(1)

            LEAK_MULTI_MODE_OFFSET = 40
            LEAK_MULTI_MODE_MASK = _mask(1)

            LEAK_ADD_MODE_OFFSET = 39
            LEAK_ADD_MODE_MASK = _mask(1)

            LEAK_TAU_OFFSET = 33
            LEAK_TAU_MASK = _mask(6)

            LEAK_V_OFFSET = 13
            LEAK_V_MASK = _mask(20)

            WEIGHT_COMPRESS_OFFSET = 12
            WEIGHT_COMPRESS_MASK = _mask(1)

            VJT_INITIAL_OFFSET = 0
            VJT_INITIAL_MASK = _mask(12)

        class Word4:
            RESET_MODE_OFFSET = 62
            RESET_MODE_MASK = _mask(2)

            RESET_V_OFFSET = 46
            RESET_V_MASK = _mask(16)

            THRESHOLD_NEG_MODE_OFFSET = 45
            THRESHOLD_NEG_MODE_MASK = _mask(1)

            THRESHOLD_POS_MODE_OFFSET = 44
            THRESHOLD_POS_MODE_MASK = _mask(1)

            THRESHOLD_NEG_OFFSET = 12
            THRESHOLD_NEG_MASK = _mask(32)

            THRESHOLD_POS_HIGH12_OFFSET = 0
            THRESHOLD_POS_HIGH12_MASK = _mask(12)

    class Fold:
        class Word1:
            FOLD_SKEW_Y_LOW2_OFFSET = 62
            FOLD_SKEW_Y_LOW2_MASK = _mask(2)

            FOLD_AXON_XY_OFFSET = 51
            FOLD_AXON_XY_MASK = _mask(11)

            FOLD_AXON_X_OFFSET = 40
            FOLD_AXON_X_MASK = _mask(11)

            FOLD_AXON_Y_OFFSET = 29
            FOLD_AXON_Y_MASK = _mask(11)

            FOLD_NUMBER_OFFSET = 0
            FOLD_NUMBER_MASK = _mask(29)

        class Word2:
            FOLD_RANGE_XY_OFFSET = 53
            FOLD_RANGE_XY_MASK = _mask(11)

            FOLD_RANGE_X_OFFSET = 42
            FOLD_RANGE_X_MASK = _mask(11)

            FOLD_RANGE_Y_OFFSET = 31
            FOLD_RANGE_Y_MASK = _mask(11)

            FOLD_SKEW_XY_OFFSET = 20
            FOLD_SKEW_XY_MASK = _mask(11)

            FOLD_SKEW_X_OFFSET = 9
            FOLD_SKEW_X_MASK = _mask(11)

            FOLD_SKEW_Y_HIGH9_OFFSET = 0
            FOLD_SKEW_Y_HIGH9_MASK = _mask(9)

        class Word3:
            FOLD_VJT_1_OFFSET = 32
            FOLD_VJT_1_MASK = _mask(32)

            FOLD_VJT_0_OFFSET = 0
            FOLD_VJT_0_MASK = _mask(32)

        class Word4:
            FOLD_VJT_3_OFFSET = 32
            FOLD_VJT_3_MASK = _mask(32)

            FOLD_VJT_2_OFFSET = 0
            FOLD_VJT_2_MASK = _mask(32)


class OfflineConfigFrame4FormatV2(FFV2):
    """Offline config frame type IV. Input SRAM."""

    pass


OfflineCoreRegFormatV2 = OfflineConfigFrame1FormatV2
OfflineLutRAMFormatV2 = OfflineConfigFrame2FormatV2
OfflineNeuRAMFormatV2 = OfflineConfigFrame3FormatV2
OfflineInutRAMFormatV2 = OfflineConfigFrame4FormatV2


class OfflineWorkFrame1FormatV2(FFV2):
    """Work frame type I. data."""

    # Time step[7]
    TIMESTEP_HIGH7_OFFSET = 60
    TIMESTEP_HIGH7_MASK = _mask(1)

    # Time step
    TIMESTEP_OFFSET = 17
    TIMESTEP_MASK = _mask(7)

    # Axon address
    AXON_ADDR_OFFSET = 8
    AXON_ADDR_MASK = _mask(9)

    # Data
    DATA_OFFSET = 0
    DATA_MASK = _mask(8)


class OfflineWorkFrame2FormatV2(FFV2):
    """Work frame type II. Vjt."""

    # Time step[7]
    TIMESTEP_HIGH7_OFFSET = 60
    TIMESTEP_HIGH7_MASK = _mask(1)

    # Time step
    TIMESTEP_LOW_OFFSET = 17
    TIMESTEP_LOW_MASK = _mask(7)

    # Axon address
    AXON_ADDR_OFFSET = 8
    AXON_ADDR_MASK = _mask(9)

    # Data Part of Vjt
    VJT_OFFSET = 0
    VJT_MASK = _mask(8)


DataFrameFormatV2 = OfflineWorkFrame1FormatV2
VjtFrameFormatV2 = OfflineWorkFrame2FormatV2


class OfflineControlFrame1FormatV2(FFV2):
    """Control frame type I. Sync."""

    # Number of time steps
    NUM_TIMESTEP_OFFSET = 0
    NUM_TIMESTEP_MASK = _mask(24)


class OfflineControlFrame2FormatV2(FFV2):
    """Control frame type II. Init."""

    # Reserved
    RESERVED_OFFSET = 0
    RESERVED_MASK = _mask(24)


class OfflineControlFrame3FormatV2(FFV2):
    """Control frame type III. Complete."""

    # THREAD_ID
    THREAD_ID_OFFSET = 0
    THREAD_ID_MASK = _mask(24)


class _OfflineTestFrameFormat_InV2(_TestFrameFormat_InV2):
    """General offline test input frame format (Read Request).

    Structure: Single Packet Header Frame.
    Common Fields:
        [23]    : Packet Type (Always 1 for Input/Request)
        [13:0]  : Number of Packets to read
    """

    TEST_PKT_TYPE_VAL = FramePackageType.TESTIN


class _OfflineTestFrameFormat_OutV2(_TestFrameFormat_OutV2):
    """General offline test output frame format (Read Response).

    Structure: Packet Header Frame + Data Body Frames.
    Common Fields:
        [23]    : Packet Type (Always 0 for Output/Response)
        [13:0]  : Number of Packets following
    """

    TEST_PKT_TYPE_VAL = FramePackageType.CONF_TESTOUT


class OfflineTestFrame1Format_InV2(_OfflineTestFrameFormat_InV2):
    """Test frame type I. Core Registers, Input (Request).

    Header Type: 0x01 (Test), Subtype: 0x00 (Core)
    Payload Layout:
        [23]   : 1'b1 (Input)
        [22:14]: Reserved
        [13:0] : Number of packets
    """

    RESERVED_OFFSET = 14
    RESERVED_MASK = _mask(9)

    PACKAGE_NUM_OFFSET = 0
    PACKAGE_NUM_MASK = _mask(14)


class OfflineTestFrame1Format_OutV2(
    _OfflineTestFrameFormat_OutV2, OfflineConfigFrame1FormatV2
):
    """Test frame type I. Core Registers, Output (Response).

    Header Type: 0x01 (Test), Subtype: 0x00 (Core)
    Payload Layout (Header):
        [23]   : 1'b0 (Output)
        [22:14]: Reserved
        [13:0] : Number of packets

    Data Body: Matches OfflineConfigFrame1FormatV2 (3 Frames: Word0~Word2).
    """

    pass


class OfflineTestFrame2Format_InV2(_OfflineTestFrameFormat_InV2):
    """Test frame type II. LUT SRAM, Input (Request).

    Header Type: 0x01 (Test), Subtype: 0x01 (LUT)
    Payload Layout:
        [23]   : 1'b1 (Input)
        [22:14]: SRAM Start Address
        [13:0] : Number of packets
    """

    SRAM_START_ADDR_OFFSET = 14
    SRAM_START_ADDR_MASK = _mask(9)

    PACKAGE_NUM_OFFSET = 0
    PACKAGE_NUM_MASK = _mask(14)


class OfflineTestFrame2Format_OutV2(
    _OfflineTestFrameFormat_OutV2, OfflineConfigFrame2FormatV2
):
    """Test frame type II. LUT SRAM, Output (Response).

    Header Type: 0x01 (Test), Subtype: 0x01 (LUT)
    Payload Layout (Header):
        [23]   : 1'b0 (Output)
        [22:14]: SRAM Start Address
        [13:0] : Number of packets

    Data Body: Matches OfflineConfigFrame2FormatV2.
    """

    pass


class OfflineTestFrame3Format_InV2(_OfflineTestFrameFormat_InV2):
    """Test frame type III. Neuron SRAM, Input (Request).

    Header Type: 0x01 (Test), Subtype: 0x10 (Neuron)
    Payload Layout:
        [23]   : 1'b1 (Input)
        [cite_start][22:14]: SRAM Start Address / 8 (Note: Scaled by 8) [cite: 71]
        [13:0] : Number of packets
    """

    SRAM_START_ADDR_DIV8_OFFSET = 14
    SRAM_START_ADDR_DIV8_MASK = _mask(9)

    PACKAGE_NUM_OFFSET = 0
    PACKAGE_NUM_MASK = _mask(14)


class OfflineTestFrame3Format_OutV2(
    _OfflineTestFrameFormat_OutV2, OfflineConfigFrame3FormatV2
):
    """Test frame type III. Neuron SRAM, Output (Response).

    Header Type: 0x01 (Test), Subtype: 0x10 (Neuron)
    Payload Layout (Header):
        [23]   : 1'b0 (Output)
        [22:14]: SRAM Start Address / 8
        [13:0] : Number of packets

    Data Body: Matches OfflineConfigFrame3FormatV2 (Full/Fold modes).
    """

    pass


class OfflineTestFrame4Format_InV2(_OfflineTestFrameFormat_InV2):
    """Test frame type IV. Input SRAM, Input (Request).

    Header Type: 0x01 (Test), Subtype: 0x11 (Input)
    Payload Layout:
        [23]   : 1'b1 (Input)
        [22:14]: SRAM Start Address
        [13:0] : Number of packets
    """

    SRAM_START_ADDR_OFFSET = 14
    SRAM_START_ADDR_MASK = _mask(9)

    PACKAGE_NUM_OFFSET = 0
    PACKAGE_NUM_MASK = _mask(14)


class OfflineTestFrame4Format_OutV2(
    _OfflineTestFrameFormat_OutV2, OfflineConfigFrame4FormatV2
):
    """Test frame type IV. Input SRAM, Output (Response).

    Header Type: 0x01 (Test), Subtype: 0x11 (Input)
    Payload Layout (Header):
        [23]   : 1'b0 (Output)
        [22:14]: SRAM Start Address
        [13:0] : Number of packets

    Data Body: Matches OfflineConfigFrame4FormatV2 (8 Words per entry).
    """

    pass


SyncFrameFormatV2 = OfflineControlFrame1FormatV2
InitFrameFormatV2 = OfflineControlFrame2FormatV2
CompleteFrameFormatv2 = OfflineControlFrame3FormatV2



from collections import UserList
from collections.abc import Sequence
from enum import Enum, IntEnum, unique
from typing import NamedTuple

from .coordinate import Coord
from .coordinate import ReplicationId as RId
from .hw_defs import HwParams

__all__ = [
    # Classes
    "RoutingCoord",
    "RoutingDirection",
    "RoutingLevel",
    "RoutingPath",
    "RoutingStatus",
    # Functions
    "get_replication_id",
    "get_multicast_cores",
    # Constants
    "ONLINE_CORES_BASE_COORD",
    "ROUTING_DIRECTIONS_IDX",
]

# The base coordinate of online cores
ONLINE_CORES_BASE_COORD = HwParams.N_CORE_OFFLINE


def _rid_unset() -> RId:
    return RId(0, 0)


@unique
class RoutingLevel(IntEnum):
    """The level of routing node.

    L0-level nodes are leaves of tree to store data. A L0-cluster is a physical core.
    """

    L0 = 0
    L1 = 1
    L2 = 2
    L3 = 3
    L4 = 4
    L5 = 5


@unique
class RoutingDirection(Enum):
    """Indicate the direction of the four children in the cluster."""

    X0Y0 = (0, 0)
    X0Y1 = (0, 1)
    X1Y0 = (1, 0)
    X1Y1 = (1, 1)
    ANY = (-1, -1)

    def to_index(self) -> int:
        """Convert the direction to index in children list."""
        if self is RoutingDirection.ANY:
            raise TypeError("The direction of routing is not specified.")

        x, y = self.value

        if HwParams.COORD_Y_PRIORITY:
            return (x << 1) + y
        else:
            return (y << 1) + x

    def __str__(self) -> str:
        if self is RoutingDirection.ANY:
            return "ANY"
        else:
            return f"X{self.value[0]}Y{self.value[1]}"


@unique
class RoutingStatus(IntEnum):
    """Indicate the status of L0-level cluster. Not used."""

    AVAILABLE = 0
    """Available for item to attach."""

    USED = 1
    """An item is attached to this cluster."""

    OCCUPIED = 2
    """Wasted. It will be an optimization goal."""

    ALL_EMPTY = 3
    """Not used."""


MAX_ROUTING_PATH_LENGTH = HwParams.N_ROUTING_PATH_LENGTH_MAX
ROUTING_DIRECTIONS_IDX = (
    [
        RoutingDirection.X0Y0,
        RoutingDirection.X0Y1,
        RoutingDirection.X1Y0,
        RoutingDirection.X1Y1,
    ]
    if HwParams.COORD_Y_PRIORITY
    else [
        RoutingDirection.X0Y0,
        RoutingDirection.X1Y0,
        RoutingDirection.X0Y1,
        RoutingDirection.X1Y1,
    ]
)


class RoutingCoord(NamedTuple):
    """Use router directions to represent the coordinate of a cluster."""

    L4: RoutingDirection = RoutingDirection.ANY
    L3: RoutingDirection = RoutingDirection.ANY
    L2: RoutingDirection = RoutingDirection.ANY
    L1: RoutingDirection = RoutingDirection.ANY
    L0: RoutingDirection = RoutingDirection.ANY

    def _coord_specify_check(self) -> None:
        if RoutingDirection.ANY in self:
            raise ValueError(
                f"the direction of routing is not specified completely, got {self}."
            )

    def _L0_property(self) -> None:
        if self.level > RoutingLevel.L0:
            raise AttributeError(
                f"this property is only for L0-level cluster, but self is {self.level}."
            )

    @property
    def level(self) -> RoutingLevel:
        for i in range(MAX_ROUTING_PATH_LENGTH):
            if self[i] is RoutingDirection.ANY:
                return RoutingLevel(MAX_ROUTING_PATH_LENGTH - i)

        return RoutingLevel.L0

    def to_coord(self) -> Coord:
        self._L0_property()
        self._coord_specify_check()

        x = sum(self[i].value[0] << (4 - i) for i in range(MAX_ROUTING_PATH_LENGTH))
        y = sum(self[i].value[1] << (4 - i) for i in range(MAX_ROUTING_PATH_LENGTH))

        return Coord(x, y)

    def __str__(self) -> str:
        return f"(L4: {self.L4}, L3: {self.L3}, L2: {self.L2}, L1: {self.L1}, L0: {self.L0})"


class RoutingPath(UserList[RoutingDirection]):
    def __init__(self, *path: RoutingDirection, reverse: bool = False) -> None:
        if reverse:
            _total_rp = [RoutingDirection.ANY] * (
                MAX_ROUTING_PATH_LENGTH - len(path)
            ) + list(reversed(path))
        else:
            _total_rp = path

        super().__init__(_total_rp)

    @property
    def routing_coord(self) -> RoutingCoord:
        return RoutingCoord(*self)

    def to_coord(self) -> Coord:
        return self.routing_coord.to_coord()

    @classmethod
    def n_core2routing_path(cls, n_core: int):
        """Return the routing path by given the #N of cores in reverse order (L4 to L0)."""
        routing_path = []

        # From L0 to L4
        for _ in range(MAX_ROUTING_PATH_LENGTH):
            n_core, re = divmod(n_core, HwParams.N_SUB_ROUTING_NODE)
            routing_path.append(ROUTING_DIRECTIONS_IDX[re])

        return cls(*routing_path, reverse=True)


def get_replication_id(coords: Sequence[Coord]) -> tuple[Coord, RId]:
    """Get the replication ID by given the coordinates.

    Args:
        - coords: sequence of coordinates.

    Returns:
        return a tuple of base coordinate & replication ID.
    """
    if len(coords) < 1:
        raise ValueError("the length of coordinates must be at least 1.")

    base_coord = coords[0]
    rid = RId(0, 0)

    for coord in coords[1:]:
        rid |= base_coord ^ coord

    base_coord_of_mcast = base_coord & (~rid)
    return base_coord_of_mcast, rid


def get_multicast_cores(base_coord: Coord, rid: RId) -> set[Coord]:
    cores = set()
    corex = set()
    corey = set()
    temp = set()

    corex.add(base_coord.x)
    corey.add(base_coord.y)

    for lx in range(5):
        if (rid.x >> lx) & 1:
            for x in corex:
                temp.add(x ^ (1 << lx))

            corex = corex.union(temp)
            temp.clear()

        if (rid.y >> lx) & 1:
            for y in corey:
                temp.add(y ^ (1 << lx))

            corey = corey.union(temp)
            temp.clear()

    for x in corex:
        for y in corey:
            cores.add(Coord(x, y))

    return cores



from abc import ABC, abstractmethod
from collections.abc import Sequence
from enum import Enum, unique
from typing import Literal, final, overload

from pydantic import Field
from pydantic.dataclasses import dataclass

from .core_defs import CoreType
from .hw_defs import HwParams, HwParamsV2
from .utils import _mask

__all__ = [
    # Classes
    "Coord",
    "OfflineCoord",
    "OnlineCoord",
    "CoordOffset",
    "ReplicationId",
    # Classes for chip v2.5
    "CoordXY",
    "CoordXYOffset",
    "CoordZXYOffset",
    "CoordXYUnitVec",
    # Aliases
    "ChipCoord",
    "CoordAddr",
    "CoordTuple2d",
    "CoordTuple3d",
    # Types
    "CoordLike",
    "RIdLike",
    "CoordXYLike",
    "CoordXYOffsetLike",
    "CoordZXYOffsetLike",
    # Functions
    "to_coord",
    "to_coords",
    "to_coordoffset",
    "to_rid",
    "to_coordxy",
    "to_coordxys",
    "to_coordxyoffset",
    "to_coordzxyoffset",
    "coordzxy_to_sign_magnitude",
]

CoordAddr = int
CoordTuple2d = tuple[int, int]
CoordTuple3d = tuple[int, int, int]

_cx_max = HwParams.CORE_X_MAX
_cx_min = HwParams.CORE_X_MIN
_cy_max = HwParams.CORE_Y_MAX
_cy_min = HwParams.CORE_Y_MIN
_cx_range = _cx_max - _cx_min + 1
_cy_range = _cy_max - _cy_min + 1


def _xy_parser(other: "CoordTuple2d | CoordOffset") -> CoordTuple2d:
    """Parse the coordinate in tuple format."""
    if not isinstance(other, (tuple, CoordOffset)):
        raise TypeError(f"unsupported type: {type(other).__name__}.")

    if isinstance(other, tuple):
        if len(other) != 2:
            raise ValueError(f"expected a tuple of 2 elements, but got {len(other)}.")

        return CoordOffset(*other).to_tuple()  # check the range of coordoffset
    else:
        return other.to_tuple()


class CoordFormat(ABC):
    """
    Identifier to descripe coordinate of hardware unit.
    """

    @abstractmethod
    def __eq__(self, other) -> bool: ...

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    @abstractmethod
    def __hash__(self): ...


class CoordVecFormat(ABC):
    """
    Identifier to descripe vector of coordinate of hardware unit.
    """

    @abstractmethod
    def __eq__(self, other) -> bool: ...

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{self.__str__()}"


UNSUPPORTED_OPERAND_TYPE_MSG = "unsupported operand type for {0}: '{1}' and '{2}'."


@dataclass
class Coord(CoordFormat):
    """Coordinate of the cores. Set coordinate (x, y) for every core.

    Left to right, +X, up to down, +Y.
    """

    x: int = Field(default=_cx_min, ge=_cx_min, le=_cx_max, description="X coordinate")
    y: int = Field(default=_cy_min, ge=_cy_min, le=_cy_max, description="Y coordinate")

    @classmethod
    def from_addr(cls, addr: CoordAddr) -> "Coord":
        if HwParams.COORD_Y_PRIORITY:
            return cls(addr >> HwParams.N_BIT_COORD_ADDR, addr & _cy_max)
        else:
            return cls(addr >> HwParams.N_BIT_COORD_ADDR, addr & _cx_max)

    def __add__(self, other: "CoordOffset") -> "Coord":
        """
        Example:
        >>> c1 = Coord(1, 1)
        >>> c2 = c1 + CoordOffset(1, 1)
        >>> c1
        >>> Coord(2, 2)

        NOTE: `Coord` + `Coord` is meaningless.
        """
        if not isinstance(other, CoordOffset):
            raise TypeError(
                UNSUPPORTED_OPERAND_TYPE_MSG.format(
                    "+", type(self).__name__, type(other).__name__
                )
            )

        sum_x, sum_y = _sum_carry(self.x + other.x, self.y + other.y)

        return Coord(sum_x, sum_y)

    def __iadd__(self, other: "CoordTuple2d | CoordOffset") -> "Coord":
        """
        Example:
        >>> c1 = Coord(1, 1)
        >>> c1 += CoordOffset(1, 1)
        >>> c1
        >>> Coord(2, 2)
        """
        bx, by = _xy_parser(other)
        self.x, self.y = _sum_carry(self.x + bx, self.y + by)

        return self

    @overload
    def __sub__(self, other: "Coord") -> "CoordOffset": ...

    @overload
    def __sub__(self, other: "CoordOffset") -> "Coord": ...

    def __sub__(self, other: "Coord | CoordOffset") -> "Coord | CoordOffset":
        """
        Example:
        >>> c1 = Coord(1, 1)
        >>> c2 = Coord(2, 2) - c1
        >>> c2
        >>> CoordOffset(1, 1)
        """
        if not isinstance(other, (Coord, CoordOffset)):
            raise TypeError(
                UNSUPPORTED_OPERAND_TYPE_MSG.format(
                    "-", type(self).__name__, type(other).__name__
                )
            )

        if isinstance(other, Coord):
            return CoordOffset(self.x - other.x, self.y - other.y)
        else:
            diff_x, diff_y = _sum_carry(self.x - other.x, self.y - other.y)
            return Coord(diff_x, diff_y)

    def __isub__(self, other: "CoordTuple2d | CoordOffset") -> "Coord":
        """
        Example:
        >>> c1 = Coord(2, 2)
        >>> c1 -= CoordOffset(1, 1)
        >>> c1
        >>> Coord(1, 1)
        """
        bx, by = _xy_parser(other)
        self.x, self.y = _sum_carry(self.x - bx, self.y - by)

        return self

    """Operations below are used only when comparing with a Cooord."""

    def __eq__(self, other: "CoordTuple2d | Coord") -> bool:
        """
        Example:
        >>> Coord(4, 5) == Coord(4, 6)
        >>> False
        """
        if not isinstance(other, (tuple, Coord)):
            raise TypeError(
                UNSUPPORTED_OPERAND_TYPE_MSG.format(
                    "==", type(self).__name__, type(other).__name__
                )
            )

        if isinstance(other, tuple):
            return self.to_tuple() == other
        else:
            return self.x == other.x and self.y == other.y

    def __and__(self, other: "Coord") -> "Coord":
        return Coord(self.x & other.x, self.y & other.y)

    def __xor__(self, other: "Coord") -> "ReplicationId":
        return ReplicationId(self.x ^ other.x, self.y ^ other.y)

    def __hash__(self) -> int:
        return hash((self.x, self.y))

    def __str__(self) -> str:
        return f"({self.x},{self.y})"

    @staticmethod
    def _to_bin(n: int, keep_bits: int) -> str:
        """Convert an integer to a binary string with a fixed number of bits, removing the prefix '0b'."""
        assert 0 <= n < (1 << keep_bits)
        return bin(n)[2:].zfill(keep_bits)

    def to_bin_str(self) -> str:
        """Convert to binary string"""
        return f"({self._to_bin(self.x, HwParams.N_BIT_COORD_ADDR)},{self._to_bin(self.y, HwParams.N_BIT_COORD_ADDR)})"

    def to_tuple(self) -> CoordTuple2d:
        """Convert to tuple"""
        return (self.x, self.y)

    def is_type_online(self) -> bool:
        """Check if the core is of online type.

        NOTE: The online core is located in the square area at the bottom right corner(+X,+Y) of the chip.
        """
        return (
            HwParams.CORE_X_ONLINE_MIN <= self.x <= HwParams.CORE_X_ONLINE_MAX
            and HwParams.CORE_Y_ONLINE_MIN <= self.y <= HwParams.CORE_Y_ONLINE_MAX
        )

    @property
    def address(self) -> CoordAddr:
        """Convert to address, 10 bits"""
        if HwParams.COORD_Y_PRIORITY:
            return (self.x << HwParams.N_BIT_COORD_ADDR) | self.y
        else:
            return (self.y << HwParams.N_BIT_COORD_ADDR) | self.x

    @property
    def core_type(self) -> CoreType:
        return CoreType.ONLINE if self.is_type_online() else CoreType.OFFLINE


ChipCoord = Coord
CoordLike = Coord | CoordAddr | CoordTuple2d


@final
class OfflineCoord(Coord):
    """Offline core coordinate"""

    x: int = Field(
        ge=HwParams.CORE_X_OFFLINE_MIN,
        le=HwParams.CORE_X_OFFLINE_MAX,
        description="X coordinate of offline core",
    )
    y: int = Field(
        ge=HwParams.CORE_Y_OFFLINE_MIN,
        le=HwParams.CORE_Y_OFFLINE_MAX,
        description="Y coordinate of offline core",
    )


@final
class OnlineCoord(Coord):
    """Online core coordinate"""

    x: int = Field(
        ge=HwParams.CORE_X_ONLINE_MIN,
        le=HwParams.CORE_X_ONLINE_MAX,
        description="X coordinate of online core",
    )
    y: int = Field(
        ge=HwParams.CORE_Y_ONLINE_MIN,
        le=HwParams.CORE_Y_ONLINE_MAX,
        description="Y coordinate of online core",
    )


@final
class ReplicationId(Coord):
    @classmethod
    def from_addr(cls, addr: CoordAddr) -> "ReplicationId":
        if HwParams.COORD_Y_PRIORITY:
            return cls(addr >> HwParams.N_BIT_COORD_ADDR, addr & _cy_max)
        else:
            return cls(addr >> HwParams.N_BIT_COORD_ADDR, addr & _cx_max)

    def __and__(self, other: "Coord | ReplicationId") -> "ReplicationId":
        return ReplicationId(self.x & other.x, self.y & other.y)

    def __or__(self, other: "Coord | ReplicationId") -> "ReplicationId":
        return ReplicationId(self.x | other.x, self.y | other.y)

    def __invert__(self) -> "ReplicationId":
        return ReplicationId(_cx_max & (~self.x), _cy_max & (~self.y))

    def __xor__(self, other: "Coord | ReplicationId") -> "ReplicationId":
        return ReplicationId(self.x ^ other.x, self.y ^ other.y)

    def __repr__(self) -> str:
        return f"RId{self.__str__()}"

    def is_type_online(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support this method"
        )

    @property
    def core_type(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support this property"
        )


RIdLike = ReplicationId | CoordAddr | CoordTuple2d


@dataclass
class CoordOffset(CoordVecFormat):
    """Offset of coordinate"""

    x: int = Field(default=_cx_min, ge=-_cx_max, le=_cx_max, description="X offset")
    y: int = Field(default=_cy_min, ge=-_cy_max, le=_cy_max, description="Y offset")

    @classmethod
    def from_offset(cls, offset: int) -> "CoordOffset":
        if HwParams.COORD_Y_PRIORITY:
            return cls(offset >> HwParams.N_BIT_COORD_ADDR, offset & _cy_max)
        else:
            return cls(offset & _cy_max, offset >> HwParams.N_BIT_COORD_ADDR)

    @overload
    def __add__(self, other: Coord) -> Coord: ...

    @overload
    def __add__(self, other: "CoordOffset") -> "CoordOffset": ...

    def __add__(self, other: "Coord | CoordOffset") -> "Coord | CoordOffset":
        """
        Examples:
        >>> delta_c1 = CoordOffset(1, 1)
        >>> delta_c2 = delta_c1 + CoordOffset(1, 1)
        >>> delta_c2
        >>> CoordOffset(2, 2)

        Coord = CoordOffset + Coord
        >>> delta_c = CoordOffset(1, 1)
        >>> c1 = Coord(2, 3)
        >>> c2 = delta_c + c1
        >>> c2
        >>> Coord(3, 4)
        """
        if isinstance(other, CoordOffset):
            # Do not carry.
            return CoordOffset(self.x + other.x, self.y + other.y)
        else:
            sum_x, sum_y = _sum_carry(self.x + other.x, self.y + other.y)
            return Coord(sum_x, sum_y)

    def __iadd__(self, other: "CoordTuple2d | CoordOffset") -> "CoordOffset":
        """
        Example:
        >>> delta_c = CoordOffset(1, 1)
        >>> delta_c += CoordOffset(1, 1)
        >>> delta_c
        >>> CoordOffset(2, 2)
        """
        bx, by = _xy_parser(other)
        self.x += bx
        self.y += by

        self._check()
        return self

    def __sub__(self, other: "CoordOffset") -> "CoordOffset":
        """
        Example:
        >>> delta_c1 = CoordOffset(1, 1)
        >>> delta_c2 = CoordOffset(2, 2)
        >>> delta_c = delta_c1 - delta_c2
        >>> delta_c
        >>> CoordOffset(-1, -1)
        """
        return CoordOffset(self.x - other.x, self.y - other.y)

    def __isub__(self, other: "CoordTuple2d | CoordOffset") -> "CoordOffset":
        """
        Example:
        >>> delta_c = CoordOffset(1, 1)
        >>> delta_c -= CoordOffset(1, 1)
        >>> delta_c
        >>> CoordOffset(0, 0)
        """
        bx, by = _xy_parser(other)
        self.x -= bx
        self.y -= by

        self._check()
        return self

    def __eq__(self, other: "CoordTuple2d | CoordOffset") -> bool:
        if isinstance(other, tuple):
            return self.to_tuple() == other
        else:
            return self.x == other.x and self.y == other.y

    def __str__(self) -> str:
        return f"({self.x},{self.y})"

    def to_tuple(self) -> CoordTuple2d:
        """Convert to tuple"""
        return (self.x, self.y)

    def _check(self) -> None:
        if (not -_cx_max <= self.x <= _cx_max) or (not -_cy_max <= self.y <= _cy_max):
            raise ValueError(
                f"offset of coordinate is out of range ({self.x}, {self.y})."
            )


_EXCEED_UPPER_TEXT = "coordinate of {0} exceeds the upper limit {1} ({2})."
_EXCEED_LOWER_TEXT = "coordinate of {0} exceeds the lower limit {1} ({2})."
_EXCEED_UPPER_CARRY_TEXT = (
    "coordinate of {0} exceeds the upper limit {1} ({2}) when {3} carries."
)
_EXCEED_LOWER_CARRY_TEXT = (
    "coordinate of {0} exceeds the lower limit {1} ({2}) when {3} carries."
)
_EXCEED_UPPER_BORROW_TEXT = (
    "coordinate of {0} exceeds the upper limit {1} ({2}) when {3} borrows."
)
_EXCEED_LOWER_BORROW_TEXT = (
    "coordinate of {0} exceeds the lower limit {1} ({2}) when {3} borrows."
)


def _sum_carry(cx: int, cy: int) -> CoordTuple2d:
    if HwParams.COORD_Y_PRIORITY:
        if cy > _cy_max:
            if cx + 1 > _cx_max:
                raise ValueError(
                    _EXCEED_UPPER_CARRY_TEXT.format("x", _cx_max, cx + 1, "y")
                )
            elif cx + 1 < _cx_min:
                raise ValueError(
                    _EXCEED_LOWER_CARRY_TEXT.format("x", _cx_min, cx + 1, "y")
                )
            else:
                cx += 1
                cy -= _cy_range
        elif _cy_min <= cy <= _cy_max:
            if cx > _cx_max:
                raise ValueError(_EXCEED_UPPER_TEXT.format("x", _cx_max, cx))
            elif cx < _cx_min:
                raise ValueError(_EXCEED_LOWER_TEXT.format("x", _cx_min, cx))
        else:  # cy < _cy_min
            if cx - 1 > _cx_max:
                raise ValueError(
                    _EXCEED_UPPER_BORROW_TEXT.format("x", _cx_max, cx - 1, "y")
                )
            elif cx - 1 < _cx_min:
                raise ValueError(
                    _EXCEED_LOWER_BORROW_TEXT.format("x", _cx_min, cx - 1, "y")
                )
            else:
                cx -= 1
                cy += _cy_range
    else:
        if cx > _cx_max:
            if cy + 1 > _cy_max:
                raise ValueError(
                    _EXCEED_UPPER_CARRY_TEXT.format("y", _cy_max, cy + 1, "x")
                )
            elif cy + 1 < _cy_min:
                raise ValueError(
                    _EXCEED_LOWER_CARRY_TEXT.format("y", _cy_min, cy + 1, "x")
                )
            else:
                cy += 1
                cx -= _cy_range
        elif _cx_min <= cx <= _cx_max:
            if cy > _cy_max:
                raise ValueError(_EXCEED_UPPER_TEXT.format("y", _cy_max, cy))
            elif cy < _cy_min:
                raise ValueError(_EXCEED_LOWER_TEXT.format("y", _cy_min, cy))
        else:  # cx < _cx_min
            if cy - 1 > _cy_max:
                raise ValueError(
                    _EXCEED_UPPER_BORROW_TEXT.format("y", _cy_max, cy - 1, "x")
                )
            elif cy - 1 < _cy_min:
                raise ValueError(
                    _EXCEED_LOWER_BORROW_TEXT.format("y", _cy_min, cy - 1, "x")
                )
            else:
                cy -= 1
                cx += _cx_range

    return cx, cy


def to_coord(coordlike: CoordLike) -> Coord:
    if isinstance(coordlike, CoordAddr):
        return Coord.from_addr(coordlike)
    elif isinstance(coordlike, tuple):
        return Coord(*coordlike)
    else:
        return coordlike


def to_coords(coordlikes: Sequence[CoordLike]) -> list[Coord]:
    return [to_coord(coordlike) for coordlike in coordlikes]


def to_coordoffset(offset: int) -> CoordOffset:
    return CoordOffset.from_offset(offset)


def to_rid(ridlike: RIdLike) -> ReplicationId:
    if isinstance(ridlike, CoordAddr):
        return ReplicationId.from_addr(ridlike)
    elif isinstance(ridlike, tuple):
        return ReplicationId(*ridlike)
    else:
        return ridlike


# Implemented for chip v2.5


@dataclass
class CoordXY(CoordFormat):
    """
    Coordinate in XY format. This describes an actual coordinate in 2d plane.

    TODO considering to deal with wraparound maps. Mirror?
    """

    x: int = Field(
        default=0,
        ge=HwParamsV2.CORE_X_MIN,
        le=HwParamsV2.CORE_X_MAX,
        description="X coordinate",
    )
    y: int = Field(
        default=0,
        ge=HwParamsV2.CORE_Y_MIN,
        le=HwParamsV2.CORE_Y_MAX,
        description="Y coordinate",
    )

    def __eq__(self, other: "CoordXY") -> bool:
        return self.x == other.x and self.y == other.y

    def __add__(self, other: "CoordXYOffset") -> "CoordXY":
        if not isinstance(other, CoordXYOffset):
            raise TypeError(
                UNSUPPORTED_OPERAND_TYPE_MSG.format(
                    "+", type(self).__name__, type(other).__name__
                )
            )

        return CoordXY(self.x + other.x, self.y + other.y)

    def __iadd__(self, other: "CoordXYOffset"):
        self.x += other.x
        self.y += other.y
        return self

    def __sub__(self, other: "CoordXY") -> "CoordXYOffset":
        if not isinstance(other, CoordXY):
            raise TypeError(
                UNSUPPORTED_OPERAND_TYPE_MSG.format(
                    "-", type(self).__name__, type(other).__name__
                )
            )

        return CoordXYOffset(self.x - other.x, self.y - other.y)

    def __isub__(self, other: "CoordXYOffset"):
        self.x -= other.x
        self.y -= other.y
        return self

    def __str__(self) -> str:
        return f"({self.x},{self.y})"

    def __hash__(self) -> int:
        return hash((self.x, self.y))

    def copy(self):
        return type(self)(self.x, self.y)

    @overload
    def neighbor(
        self, vec: "CoordXYUnitVec", inplace: Literal[True] = True
    ) -> None: ...

    @overload
    def neighbor(
        self, vec: "CoordXYUnitVec", inplace: Literal[False] = False
    ) -> "CoordXY": ...

    def neighbor(
        self, vec: "CoordXYUnitVec", inplace: bool = False
    ) -> "CoordXY | None":
        if inplace:
            self.x += vec.value.x
            self.y += vec.value.y
        else:
            return self.__add__(vec.value)


@dataclass
class CoordXYOffset(CoordVecFormat):
    """
    Offset in X, Y axis of X-Y coordinate.
    """

    x: int = Field(
        default=0,
        ge=HwParamsV2.CORE_X_MIN,
        le=HwParamsV2.CORE_X_MAX,
        description="offset in X axis",
    )
    y: int = Field(
        default=0,
        ge=HwParamsV2.CORE_Y_MIN,
        le=HwParamsV2.CORE_Y_MAX,
        description="offset in Y axis",
    )

    def __eq__(self, other: "CoordXYOffset") -> bool:
        return self.x == other.x and self.y == other.y

    def __neg__(self) -> "CoordXYOffset":
        return CoordXYOffset(-self.x, -self.y)

    def __add__(self, other: "CoordXYOffset") -> "CoordXYOffset":
        return CoordXYOffset(self.x + other.x, self.y + other.y)

    def __iadd__(self, other: "CoordXYOffset"):
        self.x += other.x
        self.y += other.y
        return self

    def __sub__(self, other: "CoordXYOffset") -> "CoordXYOffset":
        return CoordXYOffset(self.x - other.x, self.y - other.y)

    def __isub__(self, other: "CoordXYOffset"):
        self.x -= other.x
        self.y -= other.y
        return self

    def __str__(self) -> str:
        return f"({self.x},{self.y})"

    def __hash__(self) -> int:
        return hash((self.x, self.y))

    def copy(self):
        return type(self)(self.x, self.y)


@dataclass
class CoordZXYOffset(CoordVecFormat):
    """
    Offset in Z, X, Y axis of X-Y coordinate, where Z is the direction of vector (X,Y)=(1,1).
    This format is used to represent the fields 'CORE_XY', 'CORE_X' & 'CORE_Y' of the AER package.
    """

    z: int = Field(
        default=0,
        ge=HwParamsV2.CORE_Z_MIN,
        le=HwParamsV2.CORE_Z_MAX,
        description="offset in Z axis",
    )
    x: int = Field(
        default=0,
        ge=HwParamsV2.CORE_X_MIN,
        le=HwParamsV2.CORE_X_MAX,
        description="offset in X axis",
    )
    y: int = Field(
        default=0,
        ge=HwParamsV2.CORE_Y_MIN,
        le=HwParamsV2.CORE_Y_MAX,
        description="offset in Y axis",
    )

    def __eq__(self, other: "CoordZXYOffset") -> bool:
        return self.z == other.z and self.x == other.x and self.y == other.y

    def __neg__(self) -> "CoordZXYOffset":
        return CoordZXYOffset(-self.z, -self.x, -self.y)

    def __add__(self, other: "CoordZXYOffset") -> "CoordZXYOffset":
        return CoordZXYOffset(self.z + other.z, self.x + other.x, self.y + other.y)

    def __iadd__(self, other: "CoordZXYOffset"):
        self.z += other.z
        self.x += other.x
        self.y += other.y
        return self

    def __sub__(self, other: "CoordZXYOffset") -> "CoordZXYOffset":
        return CoordZXYOffset(self.z - other.z, self.x - other.x, self.y - other.y)

    def __isub__(self, other: "CoordZXYOffset"):
        self.z -= other.z
        self.x -= other.x
        self.y -= other.y
        return self

    def to_tuple(self) -> tuple[int, int, int]:
        return (self.z, self.x, self.y)

    def to_xy(self) -> CoordXYOffset:
        """Convert to X-Y offset. This means no movement in Z axis, and Z-axis offset is added to X & Y axis offset."""
        return CoordXYOffset(self.z + self.x, self.z + self.y)

    def __str__(self) -> str:
        return f"({self.z},{self.x},{self.y})"

    def __hash__(self) -> int:
        return hash(self.to_tuple())

    def copy(self):
        return type(self)(self.z, self.x, self.y)

    def to_sign_magnitude(self) -> tuple[int, int, int]:
        """Convert the ZXY coordinate to sign-magnitude format."""
        return coordzxy_to_sign_magnitude(self)


@unique
class CoordXYUnitVec(Enum):
    """
    The unit vectors of X-Y coordinate.
    NOTE: +Z = (1,1).
    """

    LOCAL = CoordXYOffset(0, 0)
    X_POS = CoordXYOffset(1, 0)
    X_NEG = -X_POS
    Y_POS = CoordXYOffset(0, 1)
    Y_NEG = -Y_POS
    Z_POS = CoordXYOffset(1, 1)
    Z_NEG = -Z_POS


CoordXYLike = CoordXY | CoordTuple2d
CoordXYOffsetLike = CoordXYOffset | CoordTuple2d
CoordZXYOffsetLike = CoordZXYOffset | CoordTuple3d


def to_coordxy(c: CoordXYLike) -> CoordXY:
    if isinstance(c, tuple):
        return CoordXY(*c)
    else:
        return c


def to_coordxys(c: Sequence[CoordXYLike]) -> list[CoordXY]:
    return [to_coordxy(c) for c in c]


def to_coordxyoffset(c: CoordXYOffsetLike) -> CoordXYOffset:
    if isinstance(c, tuple):
        return CoordXYOffset(*c)
    else:
        return c


def to_coordzxyoffset(c: CoordZXYOffsetLike) -> CoordZXYOffset:
    if isinstance(c, tuple):
        return CoordZXYOffset(*c)
    else:
        return c


def coordzxy_to_sign_magnitude(
    coordzxylike: CoordZXYOffsetLike,
) -> tuple[int, int, int]:
    """Convert the coordinate in ZXY format to sign-magnitude format."""
    if isinstance(coordzxylike, CoordZXYOffset):
        z, x, y = coordzxylike.to_tuple()
    else:
        z, x, y = coordzxylike

    def inner(n: int) -> int:
        nbits = HwParamsV2.N_BIT_COORD_ADDR
        signbit = 0 if n >= 0 else 1
        magnitude = abs(n) & _mask(nbits - 1)
        return (signbit << (nbits - 1)) | magnitude

    return (inner(z), inner(x), inner(y))



from enum import Enum, IntEnum, auto, unique
from functools import wraps
from typing import Any

from .neuron_defs import LeakComparisonMode
from .utils import _mask

__all__ = [
    # Core reg limits
    "CoreRegLim",
    "OfflineCoreRegLim",
    "OnlineCoreRegLim",
    # Types
    "WeightWidth",
    "LCN_EX",
    "InputWidthFormat",
    "SpikeWidthFormat",
    "MaxPoolingEnable",
    "SNNModeEnable",
    "LUTRandomEnable",
    "DecayRandomEnable",
    "LeakOrder",
    "OnlineModeEnable",
    # Auxiliary classes
    "CoreType",
    "CoreMode",
    # Functions
    "get_core_mode",
    "core_mode_check",
]


class CoreRegLim:
    """Limits of core registers."""

    TICK_WAIT_START_MAX = _mask(15)
    TICK_WAIT_END_MAX = _mask(15)


class OfflineCoreRegLim(CoreRegLim):
    """Limits of offline core registers."""

    pass


class OnlineCoreRegLim(CoreRegLim):
    """Limits of online core registers."""

    LATERAL_INHI_VALUE_MAX = _mask(31)
    LATERAL_INHI_VALUE_MIN = -(LATERAL_INHI_VALUE_MAX + 1)
    WEIGHT_DECAY_VALUE_MAX = _mask(7)
    WEIGHT_DECAY_VALUE_MIN = -(WEIGHT_DECAY_VALUE_MAX + 1)
    UPPER_WEIGHT_MAX = _mask(7)
    UPPER_WEIGHT_MIN = -(UPPER_WEIGHT_MAX + 1)
    LOWER_WEIGHT_MAX = UPPER_WEIGHT_MAX
    LOWER_WEIGHT_MIN = UPPER_WEIGHT_MIN
    NEU_START_MAX = _mask(10)
    NEU_END_MAX = NEU_START_MAX
    RANDOM_SEED_MAX = _mask(16)


# Type definitions of offline core registers.


@unique
class WeightWidth(IntEnum):
    """Weight bit width of crossbar. 8-bit by default."""

    WEIGHT_WIDTH_1BIT = 0
    WEIGHT_WIDTH_2BIT = 1
    WEIGHT_WIDTH_4BIT = 2
    WEIGHT_WIDTH_8BIT = 3  # Default value.


@unique
class LCN_EX(IntEnum):
    """Scale of fan-in extension."""

    LCN_1X = 0  # Default value.
    LCN_2X = 1
    LCN_4X = 2
    LCN_8X = 3
    LCN_16X = 4
    LCN_32X = 5
    LCN_64X = 6
    LCN_128X = 7  # Only for chip v2.5


@unique
class InputWidthFormat(IntEnum):
    """Format of input spike. 1-bit by default."""

    WIDTH_1BIT = 0  # Default value.
    WIDTH_8BIT = 1


@unique
class SpikeWidthFormat(IntEnum):
    """Format of output spike. 1-bit by default."""

    WIDTH_1BIT = 0  # Default value.
    WIDTH_8BIT = 1


@unique
class MaxPoolingEnable(IntEnum):
    """Enable max pooling or not in 8-bit input format. Disable by default."""

    DISABLE = 0  # Default value.
    ENABLE = 1


@unique
class SNNModeEnable(IntEnum):
    """Enable SNN mode or not. Enable by default."""

    DISABLE = 0
    ENABLE = 1  # Default value.


@unique
class CoreType(Enum):
    """Types of cores."""

    OFFLINE = auto()
    ONLINE = auto()


_ModeParamTuple = tuple[InputWidthFormat, SpikeWidthFormat, SNNModeEnable]


@unique
class CoreMode(Enum):
    """Working mode of the offline cores. Decided by `input_width`, `spike_width` and `SNN_EN` of   \
        core parameters registers.

        Mode                        input_width    spike_width    SNN_EN
        BANN                            0               0           0
        SNN                             0               0           1
        BANN/SNN to ANN                 0               1           0
        BANN/SNN to SNN with values     0               1           1
        ANN to BANN/SNN                 1               0           0
        ANN                             1               1           0
        Undefined                       1               0           1
        Undefined                       1               1           1
    """

    MODE_BANN = (
        InputWidthFormat.WIDTH_1BIT,
        SpikeWidthFormat.WIDTH_1BIT,
        SNNModeEnable.DISABLE,
    )
    MODE_SNN = (
        InputWidthFormat.WIDTH_1BIT,
        SpikeWidthFormat.WIDTH_1BIT,
        SNNModeEnable.ENABLE,
    )
    MODE_BANN_OR_SNN_TO_ANN = (
        InputWidthFormat.WIDTH_1BIT,
        SpikeWidthFormat.WIDTH_8BIT,
        SNNModeEnable.DISABLE,
    )
    MODE_BANN_OR_SNN_TO_VSNN = (
        InputWidthFormat.WIDTH_1BIT,
        SpikeWidthFormat.WIDTH_8BIT,
        SNNModeEnable.ENABLE,
    )
    MODE_ANN_TO_BANN_OR_SNN = (
        InputWidthFormat.WIDTH_8BIT,
        SpikeWidthFormat.WIDTH_1BIT,
        SNNModeEnable.DISABLE,
    )
    MODE_ANN = (
        InputWidthFormat.WIDTH_8BIT,
        SpikeWidthFormat.WIDTH_8BIT,
        SNNModeEnable.DISABLE,
    )

    @property
    def is_snn(self) -> bool:
        """Whether the SNN mode is enabled."""
        return self is CoreMode.MODE_SNN or self is CoreMode.MODE_BANN_OR_SNN_TO_VSNN

    @property
    def is_iw8(self) -> bool:
        """Wether the input width is 8-bit."""
        return self is CoreMode.MODE_ANN_TO_BANN_OR_SNN or self is CoreMode.MODE_ANN

    @property
    def is_ow8(self) -> bool:
        return (
            self is CoreMode.MODE_BANN_OR_SNN_TO_ANN
            or self is CoreMode.MODE_BANN_OR_SNN_TO_VSNN
            or self is CoreMode.MODE_ANN
        )

    @property
    def conf(self) -> _ModeParamTuple:
        return self.value


def get_core_mode(
    iw: InputWidthFormat, sw: SpikeWidthFormat, sm: SNNModeEnable
) -> CoreMode:
    try:
        return CoreMode((iw, sw, sm))
    except ValueError:
        raise ValueError(
            f"invalid mode conf: (input_width, spike_width, snn_mode) = ({iw}, {sw}, {sm}).",
        )


def core_mode_check(func):
    @wraps(func)
    def wrapper(reg_dict: dict[str, Any], *args, **kwargs):
        _ = get_core_mode(
            reg_dict["input_width"], reg_dict["spike_width"], reg_dict["snn_en"]
        )
        return func(reg_dict, *args, **kwargs)

    return wrapper


# Type definitions of online core registers.


@unique
class LUTRandomEnable(IntEnum):
    """Enable random update for LUT or not. Disable by default."""

    DISABLE = 0  # Default value.
    ENABLE = 1


@unique
class DecayRandomEnable(IntEnum):
    """Enable random update for weight decay or not. Disable by default."""

    DISABLE = 0  # Default value.
    ENABLE = 1


LeakOrder = LeakComparisonMode  # Alias, but write in core reg


@unique
class OnlineModeEnable(IntEnum):
    """Enable online mode or not (offline inference mode). Enable by default."""

    DISABLE = 0
    ENABLE = 1  # Default value.



from typing import Annotated, Any

import numpy as np
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    NonNegativeInt,
    PlainSerializer,
    PositiveInt,
    field_validator,
    model_validator,
)

from .coordinate import ChipCoord, Coord, CoordAddr, to_coord
from .core_defs import (
    LCN_EX,
    CoreRegLim,
    DecayRandomEnable,
    InputWidthFormat,
    LeakOrder,
    MaxPoolingEnable,
    OnlineCoreRegLim,
    OnlineModeEnable,
    SNNModeEnable,
    SpikeWidthFormat,
    WeightWidth,
    get_core_mode,
)
from .hw_defs import HwOfflineCoreParams as OffCoreParams
from .hw_defs import HwOnlineCoreParams as OnCoreParams
from .utils import gen_random_string

__all__ = ["CoreReg", "OfflineCoreReg", "OnlineCoreReg"]

NUM_DENDRITE_OUT_OF_RANGE_TEXT = (
    "parameter 'num_dendrite' out of range. When input width is {0}-bit, "
    + "the number of valid dendrites should be <= {1}."
)

N_REPEAT_NRAM_UNSET = 0


def _gen_random_name() -> str:
    return "CoreReg_" + gen_random_string(8)


def _lcn_le_than(v: int, le: LCN_EX) -> int:
    if v > le:
        raise ValueError(
            f"parameter 'lcn' or 'target_lcn' out of range {le.value}, but got {v}"
        )

    return v


class CoreReg(BaseModel):
    """Parameter model of registers of cores, listed in Section 2.4.1.

    NOTE: The parameters in the model are declared in `docs/Table-of-Terms.md`.
    """

    model_config = ConfigDict(
        extra="ignore", validate_assignment=True, use_enum_values=True, strict=True
    )

    name: Annotated[
        str,
        Field(
            default_factory=_gen_random_name,
            frozen=True,
            description="Name of the physical core.",
            exclude=True,
        ),
    ]


def _to_coord_check(v: Any) -> ChipCoord:
    if not isinstance(v, (Coord, CoordAddr, tuple)):
        raise TypeError(
            f"parameter 'test_chip_addr' should be a CoordLike, but got {type(v).__name__}"
        )

    return to_coord(v)


class OfflineCoreReg(CoreReg):
    """Parameter model of registers of cores, listed in Section 2.4.1.

    NOTE: The parameters in the model are declared in `docs/Table-of-Terms.md`.
    """

    weight_width: Annotated[
        WeightWidth, Field(frozen=True, description="Weight bit width of the crossbar.")
    ]

    lcn: Annotated[
        LCN_EX,
        Field(frozen=True, description="Scale of fan-in extension of the core."),
    ]

    input_width: Annotated[
        InputWidthFormat,
        Field(frozen=True, description="Width of input data."),
    ]

    spike_width: SpikeWidthFormat = Field(
        frozen=True, description="Width of output data."
    )

    num_dendrite: NonNegativeInt = Field(
        frozen=True,
        serialization_alias="neuron_num",
        description="The number of valid dendrites in the core.",
    )

    max_pooling_en: MaxPoolingEnable = Field(
        serialization_alias="pool_max",
        description="Enable max pooling or not in 8-bit input format.",
    )

    tick_wait_start: Annotated[
        NonNegativeInt,
        Field(
            frozen=True,
            le=CoreRegLim.TICK_WAIT_START_MAX,
            description="The core begins to work at #N synchronization signal. 0 for not starting   \
                while 1 for staring forever.",
        ),
    ]

    tick_wait_end: Annotated[
        NonNegativeInt,
        Field(
            frozen=True,
            le=CoreRegLim.TICK_WAIT_END_MAX,
            description="The core keeps working within #N synchronization signal. 0 for not stopping.",
        ),
    ]

    snn_en: SNNModeEnable = Field(frozen=True, description="Enable SNN mode or not.")

    target_lcn: LCN_EX = Field(
        frozen=True,
        description="The rate of the fan-in extension of the target cores.",
    )

    test_chip_addr: Annotated[
        ChipCoord,
        Field(frozen=True, description="Destination address of output test frames."),
        BeforeValidator(_to_coord_check),
        PlainSerializer(lambda v: v.address),
    ]

    n_repeat_nram: NonNegativeInt = Field(
        default=N_REPEAT_NRAM_UNSET,
        description="The number of repetitions that need to be repeated for neurons to be placed within the NRAM.",
    )

    # Add this validator since the max value of `LCN_EX` is 128X now.
    @field_validator("lcn", "target_lcn")
    @classmethod
    def lcn_limit_check(cls, v: int) -> int:
        return _lcn_le_than(v, LCN_EX.LCN_64X)

    @model_validator(mode="after")
    def dendrite_num_range_limit(self):
        _core_mode = get_core_mode(self.input_width, self.spike_width, self.snn_en)
        if _core_mode.is_iw8:
            if self.num_dendrite > OffCoreParams.N_DENDRITE_MAX_ANN:
                raise ValueError(
                    NUM_DENDRITE_OUT_OF_RANGE_TEXT.format(
                        8, OffCoreParams.N_DENDRITE_MAX_ANN
                    )
                )
        else:
            if self.num_dendrite > OffCoreParams.N_DENDRITE_MAX_SNN:
                raise ValueError(
                    NUM_DENDRITE_OUT_OF_RANGE_TEXT.format(
                        1, OffCoreParams.N_DENDRITE_MAX_SNN
                    )
                )

        return self

    @model_validator(mode="after")
    def max_pooling_disable_iw1(self):
        if (
            self.input_width == InputWidthFormat.WIDTH_1BIT
            and self.max_pooling_en == MaxPoolingEnable.ENABLE
        ):
            self.max_pooling_en = MaxPoolingEnable.DISABLE

        return self

    @model_validator(mode="after")
    def n_repeat_nram_unset(self):
        """In case 'n_repeat_nram' is unset, calculate it."""
        if self.n_repeat_nram == N_REPEAT_NRAM_UNSET:
            # Since config 'use_enum_values' is enabled, 'input_width_format' is now an integer
            # after validation.
            if self.input_width == 0:  # 1-bit
                # dendrite_comb_rate = lcn + ww
                self.n_repeat_nram = 1 << (self.lcn + self.weight_width)
            else:
                self.n_repeat_nram = 1

        return self


LUT_RANDOM_EN_LEN = OnCoreParams.LUT_LEN


def inhi_rid_range_check(rid: int) -> int:
    # 11100~11111 means the rid <= 0b00011
    if rid >= 0b100:
        raise ValueError(
            f"parameter 'inhi_core_x/y_ex' should be less than 4, but got {rid}"
        )

    return rid


def lut_random_en_check_and_convert_bitmap(v: Any) -> int:
    if not isinstance(v, (list, np.ndarray)):
        raise TypeError(
            f"parameter 'lut_random_en' should be a list or numpy array, but got {type(v).__name__}"
        )

    v_arr = np.asarray(v, dtype=bool)
    if (lut_len := v_arr.size) != LUT_RANDOM_EN_LEN:
        raise ValueError(
            f"the length of 'lut_random_en' should be {LUT_RANDOM_EN_LEN}, but got {lut_len}"
        )

    bitmap = np.sum(v_arr * (1 << np.arange(LUT_RANDOM_EN_LEN)))
    return int(bitmap)


class OnlineCoreReg(CoreReg):
    """Parameter model of registers of online cores, listed in Section 2.5.1.

    NOTE: The parameters in the model are declared in `docs/Table-of-Terms.md`.
    """

    model_config = CoreReg.model_config | {"frozen": True}

    weight_width: Annotated[
        WeightWidth,
        Field(
            serialization_alias="bit_select",
            description="Weight bit width of the crossbar.",
        ),
    ]

    lcn: Annotated[
        LCN_EX,
        Field(
            serialization_alias="group_select",
            description="Scale of fan-in extension of the core.",
        ),
    ]

    lateral_inhi_value: Annotated[
        int,
        Field(
            ge=OnlineCoreRegLim.LATERAL_INHI_VALUE_MIN,
            le=OnlineCoreRegLim.LATERAL_INHI_VALUE_MAX,
            description="The value of the lateral inhibition.",
        ),
        BeforeValidator(int),
    ]

    weight_decay_value: Annotated[
        int,
        Field(
            ge=OnlineCoreRegLim.WEIGHT_DECAY_VALUE_MIN,
            le=OnlineCoreRegLim.WEIGHT_DECAY_VALUE_MAX,
            description="The value of the weight decay.",
        ),
        BeforeValidator(int),
    ]

    upper_weight: Annotated[
        int,
        Field(
            ge=OnlineCoreRegLim.UPPER_WEIGHT_MIN,
            le=OnlineCoreRegLim.UPPER_WEIGHT_MAX,
            description="The upper limit of the weight update.",
        ),
        BeforeValidator(int),
    ]

    lower_weight: Annotated[
        int,
        Field(
            ge=OnlineCoreRegLim.LOWER_WEIGHT_MIN,
            le=OnlineCoreRegLim.LOWER_WEIGHT_MAX,
            description="The lower limit of the weight update.",
        ),
        BeforeValidator(int),
    ]

    neuron_start: Annotated[
        NonNegativeInt,
        Field(
            le=OnlineCoreRegLim.NEU_START_MAX,
            description="The start address of the valid neuron in the NRAM.",
        ),
    ]

    neuron_end: Annotated[
        NonNegativeInt,
        Field(
            le=OnlineCoreRegLim.NEU_END_MAX,
            description="The end address of the valid neuron in the NRAM.",
        ),
    ]

    inhi_core_x_ex: Annotated[
        NonNegativeInt,
        Field(
            serialization_alias="inhi_core_x_star",
            description="X replication identifier of the inhibitory core.",
        ),
        AfterValidator(inhi_rid_range_check),
    ]

    inhi_core_y_ex: Annotated[
        NonNegativeInt,
        Field(
            serialization_alias="inhi_core_y_star",
            description="Y replication identifier of the inhibitory core.",
        ),
        AfterValidator(inhi_rid_range_check),
    ]

    tick_wait_start: Annotated[
        NonNegativeInt,
        Field(
            le=CoreRegLim.TICK_WAIT_START_MAX,
            serialization_alias="core_start_time",
            description="The core begins to work at #N synchronization signal. 0 for not starting   \
                while 1 for staring forever.",
        ),
    ]

    tick_wait_end: Annotated[
        NonNegativeInt,
        Field(
            le=CoreRegLim.TICK_WAIT_END_MAX,
            serialization_alias="core_hold_time",
            description="The core keeps working within #N synchronization signal. 0 for not stopping.",
        ),
    ]

    lut_random_en: Annotated[
        int,  # bitmap of `LUT_LEN` bits
        Field(description="Enable random update for LUT or not."),
        BeforeValidator(lut_random_en_check_and_convert_bitmap),
    ]

    decay_random_en: Annotated[
        DecayRandomEnable,
        Field(description="Enable random update for weight decay or not."),
    ]

    leak_order: Annotated[
        LeakOrder,
        Field(
            serialization_alias="leakage_order",
            description="Leak after comparison or before.",
        ),
    ]

    online_mode_en: Annotated[
        OnlineModeEnable,
        Field(description="Enable online mode or not (offline inference mode)."),
    ]

    test_chip_addr: Annotated[
        ChipCoord,
        Field(
            serialization_alias="test_address",
            description="Destination address of output test frames.",
        ),
        BeforeValidator(_to_coord_check),
        PlainSerializer(lambda v: v.address),
    ]

    random_seed: Annotated[
        PositiveInt,
        Field(le=OnlineCoreRegLim.RANDOM_SEED_MAX, description="Non-zero random seed."),
    ]

    @field_validator("lcn")
    @classmethod
    def lcn_limit_check(cls, v: int) -> int:
        return _lcn_le_than(v, LCN_EX.LCN_8X)

    @model_validator(mode="after")
    def weight_update_range_check(self):
        if self.upper_weight < self.lower_weight:
            raise ValueError(
                f"parameter 'upper_weight' should be greater than or equal to 'lower_weight', but got "
                f"{self.upper_weight} < {self.lower_weight}"
            )

        return self

    @model_validator(mode="after")
    def valid_neuron_range_check(self):
        if self.neuron_start > self.neuron_end:
            raise ValueError(
                f"parameter 'neuron_start' should be less than or equal to 'neuron_end', but got "
                f"{self.neuron_start} > {self.neuron_end}"
            )

        return self



from collections import deque
from collections.abc import Generator
from enum import Flag, auto
from uuid import uuid4

from pydantic import Field
from pydantic.dataclasses import dataclass

from .coordinate import CoordXY, CoordXYUnitVec, CoordZXYOffset
from .hw_defs import HwParamsV2

__all__ = [
    "AERPacketZXYCopy",
    "AERPacket",
    "aer_packet_walk",
    "aer_packet_area",
    "find_coordxy_shortest_path",
]


@dataclass
class AERPacketZXYCopy(CoordZXYOffset):
    z: int = Field(
        default=0,
        ge=HwParamsV2.COPY_Z_MIN,
        le=HwParamsV2.COPY_Z_MAX,
        description="number of copies in Z axis",
    )
    x: int = Field(
        default=0,
        ge=HwParamsV2.COPY_X_MIN,
        le=HwParamsV2.COPY_X_MAX,
        description="number of copies in X axis",
    )
    y: int = Field(
        default=0,
        ge=HwParamsV2.COPY_Y_MIN,
        le=HwParamsV2.COPY_Y_MAX,
        description="number of copies in Y axis",
    )

    def __repr__(self) -> str:
        return f"#copy{self.__str__()}"


class PacketNextMove(Flag):
    TO_LOCAL = auto()
    TO_NEIGHBOR = auto()
    MULTICAST = TO_LOCAL | TO_NEIGHBOR


@dataclass
class AERPacket:
    """
    An AER packet consists of:
        1. the offset in (Z, X, Y) format.
        2. a number of copies in Z, X & Y axis.

    NOTE: The foothold is the coordinate where the packet is on when routing.
    """

    foothold: CoordXY = Field(
        default_factory=CoordXY, description="the coordinate where the packet is on"
    )
    offset: CoordZXYOffset = Field(
        default_factory=CoordZXYOffset,
        description="the offset of the packet in each axis",
    )
    ncopy: AERPacketZXYCopy = Field(
        default_factory=AERPacketZXYCopy,
        description="the number of copies of the packet in each axis",
    )
    id: str = Field(default_factory=lambda: uuid4().hex[:8], description="packet id")

    def __iter__(
        self,
    ) -> "Generator[tuple[PacketNextMove, AERPacket | None], None, None]":
        while True:
            if self.offset.z > 0:
                self.offset.z -= 1
                dirc = CoordXYUnitVec.Z_POS
                move = PacketNextMove.TO_NEIGHBOR
            elif self.offset.z < 0:
                self.offset.z += 1
                dirc = CoordXYUnitVec.Z_NEG
                move = PacketNextMove.TO_NEIGHBOR
            elif self.ncopy.z > 0:
                self.ncopy.z -= 1
                dirc = CoordXYUnitVec.Z_POS
                move = PacketNextMove.MULTICAST
            elif self.ncopy.z < 0:
                self.ncopy.z += 1
                dirc = CoordXYUnitVec.Z_NEG
                move = PacketNextMove.MULTICAST
            elif self.offset.x > 0:
                self.offset.x -= 1
                dirc = CoordXYUnitVec.X_POS
                move = PacketNextMove.TO_NEIGHBOR
            elif self.offset.x < 0:
                self.offset.x += 1
                dirc = CoordXYUnitVec.X_NEG
                move = PacketNextMove.TO_NEIGHBOR
            elif self.ncopy.x > 0:
                self.ncopy.x -= 1
                dirc = CoordXYUnitVec.X_POS
                move = PacketNextMove.MULTICAST
            elif self.ncopy.x < 0:
                self.ncopy.x += 1
                dirc = CoordXYUnitVec.X_NEG
                move = PacketNextMove.MULTICAST
            elif self.offset.y > 0:
                self.offset.y -= 1
                dirc = CoordXYUnitVec.Y_POS
                move = PacketNextMove.TO_NEIGHBOR
            elif self.offset.y < 0:
                self.offset.y += 1
                dirc = CoordXYUnitVec.Y_NEG
                move = PacketNextMove.TO_NEIGHBOR
            elif self.ncopy.y > 0:
                self.ncopy.y -= 1
                dirc = CoordXYUnitVec.Y_POS
                move = PacketNextMove.MULTICAST
            elif self.ncopy.y < 0:
                self.ncopy.y += 1
                dirc = CoordXYUnitVec.Y_NEG
                move = PacketNextMove.MULTICAST
            else:
                move = PacketNextMove.TO_LOCAL

            if move == PacketNextMove.MULTICAST:
                # The core will copy the packet, so change the foothold in place.
                yield move, self.neighbor(dirc)
            elif move == PacketNextMove.TO_NEIGHBOR:
                # The core will modify the foothold of the packet, so change the foothold in place.
                self.foothold.neighbor(dirc, inplace=True)
                yield move, None
            else:  # TO_LOCAL
                # The packet is arrived & stop here
                yield move, None
                break

    def arrived(self) -> bool:
        """Check if the packet is arrived."""
        return self.offset == CoordZXYOffset(
            0, 0, 0
        ) and self.ncopy == AERPacketZXYCopy(0, 0, 0)

    def neighbor(self, direction: CoordXYUnitVec):
        """Move the packet to its neighbor at the given direction."""
        next_foothold = self.foothold.neighbor(direction, inplace=False)
        return type(self)(next_foothold, self.offset.copy(), self.ncopy.copy())

    def copy(self):
        return type(self)(self.foothold.copy(), self.offset.copy(), self.ncopy.copy())


def aer_packet_walk(packet: AERPacket) -> list[CoordXY]:
    """
    Simulate the AER packet routing behavior of chip v2.5.

    Returns:
        A list of coordinates that the packet passed through.
    """
    queue = deque([packet])
    cover_coords = list()  # ordered
    visited = set()

    while queue:
        p = queue.popleft()
        for move, new_pkg in p:
            cur_foot = p.foothold.copy()
            if PacketNextMove.TO_LOCAL in move:
                # TO_LOCAL or MULTICAST
                if cur_foot not in visited:  # walk to a new coordinate
                    visited.add(cur_foot)
                    cover_coords.append(cur_foot)

            if move == PacketNextMove.MULTICAST:
                assert new_pkg is not None
                queue.append(new_pkg)

    return cover_coords


def aer_packet_area(ncopy: AERPacketZXYCopy | tuple[int, int, int]) -> int:
    """
    Get the area of the AER packet. The area is only dependent on the 'ncopy'.
    """
    if isinstance(ncopy, tuple):
        z, x, y = map(abs, ncopy)
    else:
        z, x, y = abs(ncopy.z), abs(ncopy.x), abs(ncopy.y)

    return 1 + z + (1 + z) * x + (1 + z + x) * y


def find_coordxy_shortest_path(
    target: CoordXY, start: CoordXY = CoordXY(0, 0)
) -> tuple[CoordZXYOffset, int]:
    """
    Find the shortest path(in ZXY offset format) from `start` to `target`, and return the length.

    vec = (dx, dy) = (x2 - x1, y2 - y1)
    Assuming that z(1, 1) + x(1, 0) + y(0, 1) = vec & min(|z| + |x| + |y|)
    --> z + x = dx
        z + y = dy
    --> x = dx - z, y = dy - z
    --> min(|z| + |x| + |y|) = min(|z| + |dx - z| + |dy - z|)
    --> o1 = dx, o2 = dy
    """
    vec = target - start
    dx, dy = vec.x, vec.y
    zeros = [0, dx, dy]

    best_z = sorted(zeros)[1]
    best_x = dx - best_z
    best_y = dy - best_z
    min_cost = abs(best_z) + abs(best_x) + abs(best_y)

    return CoordZXYOffset(best_z, best_x, best_y), min_cost



from typing import TypeVar

import numpy as np
from numpy.typing import NDArray

FRAME_DTYPE = np.uint64
PAYLOAD_DATA_DTYPE = np.uint8  # in work frame type I
LUT_DTYPE = np.int8
FrameArrayType = NDArray[FRAME_DTYPE]
PayloadDataType = NDArray[PAYLOAD_DATA_DTYPE]
LUTDataType = NDArray[LUT_DTYPE]

LUT_POTENTIAL_DTYPE = np.int32
LUTPotentialType = NDArray[LUT_POTENTIAL_DTYPE]
LUT_ACTIVATION_DTYPE = np.int8 | np.uint8
LUTActivationType = NDArray[LUT_ACTIVATION_DTYPE]

FrameArrayLike = TypeVar(
    "FrameArrayLike", int, list[int], tuple[int, ...], NDArray[FRAME_DTYPE]
)
IntScalarType = TypeVar("IntScalarType", int, bool, np.integer)
DataType = TypeVar("DataType", int, bool, np.integer, np.ndarray)



from enum import Enum, unique

from .hw_defs import HwOfflineCoreParams as HwOffCore
from .hw_defs import HwOnlineCoreParams as HwOnCore
from .hw_defs import HwParams
from .utils import _mask

__all__ = [
    # Neuron reg limits
    "NeuRegLim",
    "OfflineNeuRegLim",
    "OnlineNeuRegLim",
    "OnlineNeuRegLim_WW1",
    "OnlineNeuRegLim_WWn",
    # Types
    "ResetMode",
    "LeakComparisonMode",
    "NegativeThresholdMode",
    "LeakDirectionMode",
    "LeakIntegrationMode",
    "SynapticIntegrationMode",
]


class NeuRegLim:
    """Limits of neuron registers."""

    ADDR_COORD_MAX = _mask(HwParams.N_BIT_COORD_ADDR)


class OfflineNeuRegLim(NeuRegLim):
    """Limits of offline neuron registers."""

    ADDR_TS_MAX = HwOffCore.N_TIMESLOT_MAX - 1
    ADDR_AXON_MAX = HwOffCore.ADDR_AXON_MAX
    RESET_V_MAX = _mask(29)
    RESET_V_MIN = -(RESET_V_MAX + 1)
    THRES_MASK_BITS_MAX = _mask(5)
    NEG_THRES_MAX = _mask(29)
    POS_THRES_MAX = _mask(29)
    LEAK_V_MAX = _mask(29)
    LEAK_V_MIN = -(LEAK_V_MAX + 1)
    BIT_TRUNC_MAX = 29  # The highest bit is the sign bit
    VOLTAGE_MAX = _mask(29)
    VOLTAGE_MIN = -(VOLTAGE_MAX + 1)


class OnlineNeuRegLim(NeuRegLim):
    """Basic limits of online neuron registers."""

    ADDR_TS_MAX = HwOnCore.N_TIMESLOT_MAX - 1
    ADDR_AXON_MAX = HwOnCore.ADDR_AXON_MAX
    PLASTICITY_START_MAX = _mask(10)  # <= ADDR_AXON_MAX
    PLASTICITY_END_MAX = PLASTICITY_START_MAX  # <= ADDR_AXON_MAX


class OnlineNeuRegLim_WW1(OnlineNeuRegLim):
    """Limits of online neuron registers with 1-bit weight width."""

    LEAK_V_MAX = _mask(14)
    LEAK_V_MIN = -(LEAK_V_MAX + 1)
    THRES_MAX = _mask(14)
    THRES_MIN = -(THRES_MAX + 1)
    FLOOR_THRES_MAX = _mask(6)
    FLOOR_THRES_MIN = -(FLOOR_THRES_MAX + 1)
    RESET_V_MAX = _mask(5)
    RESET_V_MIN = -(RESET_V_MAX + 1)
    INIT_V_MAX = _mask(5)
    INIT_V_MIN = -(INIT_V_MAX + 1)
    VOLTAGE_MAX = _mask(14)
    VOLTAGE_MIN = -(VOLTAGE_MAX + 1)


class OnlineNeuRegLim_WWn(OnlineNeuRegLim):
    """Limits of online neuron registers with weight width > 1-bit."""

    LEAK_V_MAX = _mask(31)
    LEAK_V_MIN = -(LEAK_V_MAX + 1)
    THRES_MAX = _mask(31)
    THRES_MIN = -(THRES_MAX + 1)
    FLOOR_THRES_MAX = _mask(31)
    FLOOR_THRES_MIN = -(FLOOR_THRES_MAX + 1)
    RESET_V_MAX = _mask(31)
    RESET_V_MIN = -(RESET_V_MAX + 1)
    INIT_V_MAX = _mask(31)
    INIT_V_MIN = -(INIT_V_MAX + 1)
    VOLTAGE_MAX = _mask(31)
    VOLTAGE_MIN = -(VOLTAGE_MAX + 1)


@unique
class ResetMode(Enum):
    """Reset modes of neurons. 2-bit.
    - `MODE_NORMAL`: normal mode. Default value.
    - `MODE_LINEAR`: linear mode.
    - `MODE_NONRESET`: non-reset mode.
    """

    MODE_NORMAL = 0  # Default value.
    MODE_LINEAR = 1
    MODE_NONRESET = 2


@unique
class LeakComparisonMode(Enum):
    """Leak after comparison or before. Default is `LEAK_BEFORE_COMP`."""

    LEAK_BEFORE_COMP = 0  # Default value.
    LEAK_AFTER_COMP = 1


@unique
class NegativeThresholdMode(Enum):
    """Modes of negative threshold. 1-bit.

    - `MODE_RESET`: reset mode. Default value.
    - `MODE_SATURATION`: saturation(floor) mode.

    NOTE: Same as `threshold_neg_mode` in V2.1.
    """

    MODE_RESET = 0  # Default value.
    MODE_SATURATION = 1


@unique
class LeakDirectionMode(Enum):
    """Direction of Leak, forward or reversal.

    - `MODE_FORWARD`: forward Leak. Default value.
    - `MODE_REVERSAL`: reversal Leak.

    NOTE: Same as `leak_reversal_flag` in V2.1.
    """

    MODE_FORWARD = 0  # Default value.
    MODE_REVERSAL = 1


@unique
class LeakIntegrationMode(Enum):
    """Mode of Leak integration, deterministic or stochastic.

    - `MODE_DETERMINISTIC`: deterministic Leak. Default value.
    - `MODE_STOCHASTIC`: stochastic Leak.

    NOTE: Same as `leak_det_stoch` in V2.1.
    """

    MODE_DETERMINISTIC = 0  # Default value.
    MODE_STOCHASTIC = 1


@unique
class SynapticIntegrationMode(Enum):
    """Modes of synaptic integration, deterministic or stochastic.

    - `MODE_DETERMINISTIC`: deterministic weights. Default value.
    - `MODE_STOCHASTIC`: stochastic weights.

    NOTE: Same as `weight_det_stoch` in V2.1.
    """

    MODE_DETERMINISTIC = 0  # Default value.
    MODE_STOCHASTIC = 1



import operator
import warnings
from typing import Annotated, TypeVar

import numpy as np
from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    InstanceOf,
    NonNegativeInt,
    PlainSerializer,
    TypeAdapter,
    ValidationInfo,
    field_validator,
    model_validator,
)

# Use `typing_extensions.TypedDict`
from typing_extensions import TypedDict

from .core_defs import WeightWidth
from .neuron_defs import LeakComparisonMode as LCMode
from .neuron_defs import LeakDirectionMode as LDMode
from .neuron_defs import LeakIntegrationMode as LIMode
from .neuron_defs import NegativeThresholdMode as NTMode
from .neuron_defs import (
    NeuRegLim,
    OfflineNeuRegLim,
    OnlineNeuRegLim,
    OnlineNeuRegLim_WW1,
    OnlineNeuRegLim_WWn,
    ResetMode,
)
from .neuron_defs import SynapticIntegrationMode as SIMode
from .utils import ndarray_serializer, range_check_unsigned

__all__ = [
    "NeuDestInfo",
    "OfflineNeuDestInfo",
    "OnlineNeuDestInfo",
    "NeuAttrs",
    "OfflineNeuAttrs",
    "OnlineNeuAttrs",
    "OfflineNeuConf",
    "OnlineNeuConf",
]


class NeuDestInfo(BaseModel):
    model_config = ConfigDict(extra="ignore", frozen=True, strict=True)

    addr_chip_x: Annotated[
        NonNegativeInt,
        Field(
            le=NeuRegLim.ADDR_COORD_MAX,
            description="X coordinate of the target chip of the neuron.",
        ),
    ]

    addr_chip_y: Annotated[
        NonNegativeInt,
        Field(
            le=NeuRegLim.ADDR_COORD_MAX,
            description="Y coordinate of the target chip of the neuron.",
        ),
    ]

    addr_core_x: Annotated[
        NonNegativeInt,
        Field(
            le=NeuRegLim.ADDR_COORD_MAX,
            description="X coordinate of the target core of the neuron.",
        ),
    ]

    addr_core_y: Annotated[
        NonNegativeInt,
        Field(
            le=NeuRegLim.ADDR_COORD_MAX,
            description="Y coordinate of the target core of the neuron.",
        ),
    ]

    addr_core_x_ex: Annotated[
        NonNegativeInt,
        Field(
            le=NeuRegLim.ADDR_COORD_MAX,
            description="X replication identifier of the neuron.",
        ),
    ]

    addr_core_y_ex: Annotated[
        NonNegativeInt,
        Field(
            le=NeuRegLim.ADDR_COORD_MAX,
            description="Y replication identifier of the neuron.",
        ),
    ]

    tick_relative: Annotated[
        list[InstanceOf[NonNegativeInt]], Field(description="Destination timeslot.")
    ]

    addr_axon: Annotated[
        list[InstanceOf[NonNegativeInt]], Field(description="Destination axon address.")
    ]

    @model_validator(mode="after")
    def length_match_check(self):
        if len(self.tick_relative) != len(self.addr_axon):
            raise ValueError(
                "parameter 'tick_relative' & 'addr_axon' must have the same length, "
                f"but {len(self.tick_relative)} != {len(self.addr_axon)}."
            )

        return self


class OfflineNeuDestInfo(NeuDestInfo):
    @field_validator("tick_relative")
    @classmethod
    def tick_relative_check(cls, v):
        return range_check_unsigned(v, "tick_relative", OfflineNeuRegLim.ADDR_TS_MAX)

    @field_validator("addr_axon")
    @classmethod
    def addr_axon_check(cls, v):
        # NOTE: When offline core -> online core, the upper limit of `addr_axon` is `OnlineNeuRegLim.ADDR_AXON_MAX`.
        # Use `max` to cover both cases.
        return range_check_unsigned(
            v,
            "addr_axon",
            max(OfflineNeuRegLim.ADDR_AXON_MAX, OnlineNeuRegLim.ADDR_AXON_MAX),
        )


class OnlineNeuDestInfo(NeuDestInfo):
    @field_validator("tick_relative")
    @classmethod
    def tick_relative_check(cls, v):
        return range_check_unsigned(v, "tick_relative", OnlineNeuRegLim.ADDR_TS_MAX)

    @field_validator("addr_axon")
    @classmethod
    def addr_axon_check(cls, v):
        # NOTE: When online core -> offline core, the upper limit of `addr_axon` is `OfflineNeuRegLim.ADDR_AXON_MAX`
        # Use `max` to cover both cases.
        return range_check_unsigned(
            v,
            "addr_axon",
            max(OfflineNeuRegLim.ADDR_AXON_MAX, OnlineNeuRegLim.ADDR_AXON_MAX),
        )


class NeuAttrs(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        arbitrary_types_allowed=True,
        use_enum_values=True,
        strict=True,
    )


class OfflineNeuAttrs(NeuAttrs):
    reset_mode: Annotated[ResetMode, Field(description="Reset mode of neuron.")]

    reset_v: Annotated[
        int,
        Field(
            ge=OfflineNeuRegLim.RESET_V_MIN,
            le=OfflineNeuRegLim.RESET_V_MAX,
            description="Reset voltage, 30-bit signed integer.",
        ),
    ]

    leak_comparison: Annotated[
        LCMode,
        Field(
            serialization_alias="leak_post",
            description="Leak after threshold comparison or before.",
        ),
    ]

    thres_mask_bits: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLim.THRES_MASK_BITS_MAX,
            serialization_alias="threshold_mask_ctrl",
            description="Bit mask for random threshold.",
        ),
    ]

    neg_thres_mode: Annotated[
        NTMode,
        Field(
            serialization_alias="threshold_neg_mode",
            description="Mode of negative threshold.",
        ),
    ]

    neg_threshold: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLim.NEG_THRES_MAX,
            serialization_alias="threshold_neg",
            description="Negative threshold, 29-bit unsigned integer.",
        ),
    ]

    pos_threshold: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLim.POS_THRES_MAX,
            serialization_alias="threshold_pos",
            description="Positive threshold, 29-bit unsigned integer.",
        ),
    ]

    leak_direction: Annotated[
        LDMode,
        Field(
            serialization_alias="leak_reversal_flag",
            description="Direction of leak, forward or reversal.",
        ),
    ]

    leak_integration_mode: Annotated[
        LIMode,
        Field(
            serialization_alias="leak_det_stoch",
            description="Mode of leak integration, deterministic or stochastic.",
        ),
    ]

    leak_v: Annotated[
        int | np.ndarray,
        Field(description="Leak voltage, 30-bit signed integer or a np.int32 array."),
        PlainSerializer(ndarray_serializer, when_used="json"),
    ]

    syn_integration_mode: Annotated[
        SIMode,
        Field(
            serialization_alias="weight_det_stoch",
            description="Mode of synaptic integration, deterministic or stochastic.",
        ),
    ]

    bit_trunc: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLim.BIT_TRUNC_MAX,
            serialization_alias="bit_truncate",
            description="Position of truncation, 5-bit unsigned integer.",
        ),
    ]

    voltage: Annotated[
        int,
        Field(
            default=0,
            description="Initial voltage, 30-bit signed integer. Fixed at 0 at initialization.",
        ),
        BeforeValidator(int),
    ] = 0


class OfflineNeuConf(BaseModel):
    attrs: OfflineNeuAttrs
    dest_info: OfflineNeuDestInfo


val_range_name = {
    "leak_v": ("LEAK_V_MIN", "LEAK_V_MAX"),
    "pos_threshold": ("THRES_MIN", "THRES_MAX"),
    "neg_threshold": ("FLOOR_THRES_MIN", "FLOOR_THRES_MAX"),
    "reset_v": ("RESET_V_MIN", "RESET_V_MAX"),
    "init_v": ("INIT_V_MIN", "INIT_V_MAX"),
    "voltage": ("VOLTAGE_MIN", "VOLTAGE_MAX"),
}

_VT = TypeVar("_VT", int, np.ndarray)


def _validate_range(field: str, value: _VT, info: ValidationInfo) -> _VT:
    if info.context is None:
        warnings.warn(
            "context 'weight_width' is not provided. Assuming 8-bit weight width.",
            UserWarning,
        )
        ww = WeightWidth.WEIGHT_WIDTH_8BIT
    else:
        assert isinstance(
            info.context, dict
        ), f"context must be dict, but got {type(info.context).__name__}."
        if "weight_width" not in info.context:
            warnings.warn(
                "context 'weight_width' is not provided. Assuming 8-bit weight width.",
                UserWarning,
            )
            ww = WeightWidth.WEIGHT_WIDTH_8BIT
        else:
            ww = info.context["weight_width"]
            if not isinstance(ww, WeightWidth):
                raise TypeError(
                    f"context 'weight_width' must be of type 'WeightWidth', but got {type(ww).__name__}."
                )

    if field not in val_range_name:
        raise ValueError(f"invalid field name: {field}")

    min_attr, max_attr = val_range_name[field]
    getter = operator.attrgetter(min_attr, max_attr)

    if ww == WeightWidth.WEIGHT_WIDTH_1BIT:
        min_val, max_val = getter(OnlineNeuRegLim_WW1)
    else:
        min_val, max_val = getter(OnlineNeuRegLim_WWn)

    if isinstance(value, int):
        if not (min_val <= value <= max_val):
            raise ValueError(f"parameter '{field}' out of range [{min_val}, {max_val}]")
    else:
        _min, _max = np.min(value), np.max(value)
        if not (min_val <= _min <= _max <= max_val):
            raise ValueError(f"parameter '{field}' out of range [{min_val}, {max_val}]")

        if field in ("leak_v", "init_v"):
            value = value.astype(np.int32)

    return value


class OnlineNeuAttrs(NeuAttrs):
    leak_v: Annotated[
        int | np.ndarray,
        Field(
            serialization_alias="leakage_reg",
            description="Leak voltage, 15-/32-bit signed integer or a np.int32 array.",
        ),
        PlainSerializer(ndarray_serializer, when_used="json"),
    ]

    pos_threshold: Annotated[
        int,
        Field(
            serialization_alias="threshold_reg",
            description="Pos threshold, 15-/32-bit signed integer.",
        ),
    ]

    neg_threshold: Annotated[
        int,
        Field(
            serialization_alias="floor_threshold_reg",
            description="Floor threshold, 7-/32-bit signed integer.",
        ),
    ]

    reset_v: Annotated[
        int,
        Field(
            serialization_alias="reset_potential_reg",
            description="Reset voltage, 6-/32-bit signed integer.",
        ),
    ]

    init_v: Annotated[
        int | np.ndarray,
        Field(
            default=0,
            serialization_alias="initital_potential_reg",
            description="Initial voltage, 6-/32-bit signed integer.",
        ),
        PlainSerializer(ndarray_serializer, when_used="json"),
    ] = 0

    voltage: Annotated[
        int,
        Field(
            default=0,
            serialization_alias="potential_reg",
            description="Initial voltage, 15-/32-bit signed integer. Fixed at 0.",
        ),
        BeforeValidator(int),
    ] = 0

    plasticity_start: Annotated[
        NonNegativeInt,
        Field(
            default=0,
            le=OnlineNeuRegLim.PLASTICITY_START_MAX,
            description="Position where plasticity starts.",
        ),
    ] = 0

    plasticity_end: Annotated[
        NonNegativeInt,
        Field(
            default=OnlineNeuRegLim.PLASTICITY_END_MAX,
            le=OnlineNeuRegLim.PLASTICITY_END_MAX,
            description="Position where plasticity ends.",
        ),
    ]

    @field_validator(*["leak_v", "reset_v", "pos_threshold", "neg_threshold", "init_v"])
    @classmethod
    def validate_ranges(cls, v: _VT, info: ValidationInfo) -> _VT:
        if info.field_name is None:
            raise ValueError("'field_name' is None")

        return _validate_range(info.field_name, v, info)

    @model_validator(mode="after")
    def plasticity_range_check(self):
        if self.plasticity_start > self.plasticity_end:
            raise ValueError(
                "'plasticity_start' must be less than or equal to 'plasticity_end', but got "
                f"{self.plasticity_start} > {self.plasticity_end}"
            )

        return self


class OnlineNeuConf(BaseModel):
    attrs: OnlineNeuAttrs
    dest_info: OnlineNeuDestInfo


class NeuDestInfoDict(TypedDict):
    """Typed dictionary of `NeuronDestInfo` for typing check."""

    addr_core_x: NonNegativeInt
    addr_core_y: NonNegativeInt
    addr_core_x_ex: NonNegativeInt
    addr_core_y_ex: NonNegativeInt
    addr_chip_x: NonNegativeInt
    addr_chip_y: NonNegativeInt
    tick_relative: list[InstanceOf[NonNegativeInt]]
    addr_axon: list[InstanceOf[NonNegativeInt]]


OfflineNeuDestInfoChecker = TypeAdapter(OfflineNeuDestInfo)
OnlineNeuDestInfoChecker = TypeAdapter(OnlineNeuDestInfo)



from typing import Annotated

from pydantic import Field, NonNegativeInt, PositiveInt, model_validator

from .core_defs import LCN_EX  # Reuse some types
from .core_defs_v2 import (
    AddPotentialMode,
    CSCAccelerateMode,
    DataSign,
    DataWidth,
    OfflineCoreRegLimV2,
    PoolingMode,
    SNNMode,
    ZeroOutputMode,
)
from .core_model import CoreReg

__all__ = ["OfflineCoreRegV2"]


class OfflineCoreRegV2(CoreReg):
    """Core parameters model, corresponding to Section 2.3.1 Core Parameters."""

    # Configuration Parameters
    snn_ann: Annotated[SNNMode, Field(description="SNN and ANN mode selection.")]
    max_pooling: Annotated[PoolingMode, Field(description="Pooling mode selection.")]
    add_potential: Annotated[
        AddPotentialMode, Field(description="Accumulation mode selection.")
    ]
    zero_output: Annotated[
        ZeroOutputMode, Field(description="Whether to output zero values.")
    ]

    input_sign: Annotated[DataSign, Field(description="Input data sign selection.")]
    input_width: Annotated[
        DataWidth, Field(description="Input data bit width selection.")
    ]

    output_sign: Annotated[DataSign, Field(description="Output data sign selection.")]
    output_width: Annotated[
        DataWidth, Field(description="Output data bit width selection.")
    ]

    weight_sign: Annotated[DataSign, Field(description="Weight data sign selection.")]
    weight_width: Annotated[
        DataWidth, Field(description="Weight data bit width selection.")
    ]

    lcn: Annotated[LCN_EX, Field(description="Control the scale of fan-in extension.")]
    target_lcn: Annotated[
        LCN_EX, Field(description="LCN of the output target address core.")
    ]

    axon_skew: Annotated[
        int,
        Field(
            ge=OfflineCoreRegLimV2.AXON_SKEW_MIN,
            le=OfflineCoreRegLimV2.AXON_SKEW_MAX,
            description="Axon address offset for AER format input work frame.",
        ),
    ]
    neuron_number: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineCoreRegLimV2.NEURON_NUMBER_MAX,
            description="Number of valid neuron addresses.",
        ),
    ]

    # Test/Control Frame Addresses (Sign-Magnitude 6-bit: -31 to 31)
    test_core_xy: Annotated[
        int,
        Field(
            ge=OfflineCoreRegLimV2.TEST_CORE_COORD_MIN,
            le=OfflineCoreRegLimV2.TEST_CORE_COORD_MAX,
            description="Relative XY address of the core sent by test/control frame.",
        ),
    ]
    test_core_x: Annotated[
        int,
        Field(
            ge=OfflineCoreRegLimV2.TEST_CORE_COORD_MIN,
            le=OfflineCoreRegLimV2.TEST_CORE_COORD_MAX,
            description="Relative X address of the core sent by test/control frame.",
        ),
    ]
    test_core_y: Annotated[
        int,
        Field(
            ge=OfflineCoreRegLimV2.TEST_CORE_COORD_MIN,
            le=OfflineCoreRegLimV2.TEST_CORE_COORD_MAX,
            description="Relative Y address of the core sent by test/control frame.",
        ),
    ]

    # Global Signals
    global_send: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineCoreRegLimV2.GLOBAL_SEND_MAX,
            description="Global signal send direction (local, xy+, xy-, x+, x-, y+, y-).",
        ),
    ]

    csc_accelerate: Annotated[
        CSCAccelerateMode,
        Field(description="CSC compressed calculation acceleration mode."),
    ]

    # Global Signals
    global_receive: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineCoreRegLimV2.GLOBAL_RECEIVE_MAX,
            description="Global signal receive direction (xy+, xy-, x+, x-, y+, y-).",
        ),
    ]

    # Thread and Timing
    thread_number: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineCoreRegLimV2.THREAD_NUMBER_MAX,
            description="Thread number of the current core.",
        ),
    ]
    busy_cycle: Annotated[
        PositiveInt,
        Field(
            gt=1,
            le=OfflineCoreRegLimV2.BUSY_CYCLE_MAX,
            description="Mask threshold for busy signal, usually >1.",
        ),
    ]
    delay_cycle: Annotated[
        PositiveInt,
        Field(
            gt=1,
            le=OfflineCoreRegLimV2.DELAY_CYCLE_MAX,
            description="Delay time for control signal to take effect, usually >1.",
        ),
    ]
    width_cycle: Annotated[
        PositiveInt,
        Field(
            gt=1,
            le=OfflineCoreRegLimV2.WIDTH_CYCLE_MAX,
            description="Multi-cycle width for control global signals sync_all, initial_all, usually >1.",
        ),
    ]

    # Tick Control
    tick_start: Annotated[
        NonNegativeInt,
        Field(le=OfflineCoreRegLimV2.TICK_START_MAX, description="Start tick count."),
    ]
    tick_duration: Annotated[
        NonNegativeInt,
        Field(
            default=0,
            le=OfflineCoreRegLimV2.TICK_DURATION_MAX,
            description="Duration tick count.",
        ),
    ] = 0
    tick_initial: Annotated[
        NonNegativeInt,
        Field(
            default=0,
            le=OfflineCoreRegLimV2.TICK_INITIAL_MAX,
            description="Auto-initialization tick count.",
        ),
    ] = 0

    @model_validator(mode="after")
    def check_weight_width(self):
        if (
            self.max_pooling == PoolingMode.MAX
            or self.add_potential == AddPotentialMode.DIRECT_ADD
        ):
            if self.weight_width != DataWidth.WIDTH_1BIT:
                self.weight_width = DataWidth.WIDTH_1BIT

        return self
